/*
 *  $Header: /auto/nfs/sequoia/u/an/work/rws/pextk/src/pextk_buf.c,v 1.3 1994/01/04 01:18:58 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                          pextk_buf.c                               *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *   written by: Bob Schulte		date: February 1992                   *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk buffering routines.
 *
 *  Module Notes:
 *
 *  Module Contents:
 *
 *    int   PEXtk_init_buf          Initializes the buffering routines
 *    void  PEXtk_copy_window       Swaps buffers using copy area
 *    int   PEXtk_create_back_pixmap Creates the back buffer using pixmaps
 *    int   PEXtk_delete_back_pixmap Deletes the back buffer using pixmaps
 *    int   PEXtk_create_mbuf       Creates the back buffer using Multi-Buf
 *    int   PEXtk_delete_mbuf       Deletes the back buffer using Multi-Buf
 *    int   PEXtk_copy_window       Swaps buffer using XCopyArea
 *    int   PEXtk_swap_mbuf         Swaps buffer using Multi-Buf
 *    int   PEXtk_clear_area        Clears the image plane using XClearArea
 *    int   PEXtk_clear_iz_5        Clears the image,Z planes using XFillRect
 *    int   PEXtk_clear_i_5         Clears the image plane using XFillRect
 *    int   PEXtk_clear_z_5         Clears zbuffer using begin/end render
 *    int   PEXtk_clear_none        Performs no z or image clear
 *    int   PEXtk_clear_iz          Clears the image,Z planes using cleari,z
 *    int   PEXtk_clear_i           Clears the image plane using clear_image
 *    int   PEXtk_clear_z           Clears zbuffer using begin/end render
 *    int   PEXtk_mpex_clear_z      Clears zbuffer using MPEX clearZ
 *    int   PEXtk_mpex_clear_i      Clears image using MPEX clearI.
 *    int   PEXtk_mpex_clear_iz     Clears image,Z using MPEX clearI and clearZ
 *
 *  Module Description:
 *	This module handles the buffering capabilities of PEXtk.
 *
 *	PEXtk provides support for the specification of a user supplied
 *	routine to perform the various back buffer manipulation and
 *	pixel lookup.
 *
 *	To replace the PEXtk routine with another routine, call the
 *	Pset_function routine with 'type' set to the function type to
 *	replace, and pass in the address of the function.
 *
 *	The following is a table for the 'Pset_function' type and the
 *	associated PEXtk default routine.
 *
 *	  	Type		      default		      description
 *    -----------------------------------------------------------------------
 *    PEXtk_FUNC_SWAP_BUFFER	  PEXtk_swap_mbuf	    MBX buffer swap
 *    PEXtk_FUNC_SWAP_BUFFER	  PEXtk_copy_window	    Non-MBX buffer swap
 *    PEXtk_FUNC_CREATE_BUFFER    PEXtk_create_mbuf	    MBX buffer create
 *    PEXtk_FUNC_CREATE_BUFFER    PEXtk_create_back_pixmap  Non-MBX 
 *    PEXtk_FUNC_DELETE_BUFFER    PEXtk_delete_mbuf	    MBX buffer delete
 *    PEXtk_FUNC_DELETE_BUFFER    PEXtk_delete_back_pixmap  Non-MBX 
 *    PEXtk_FUNC_DELETE_BUFFER    PEXtk_delete_back_pixmap  Non-MBX 
 *
 *    PEXtk_FUNC_PIXEL_RGB    	  Ppseudo_color  	 8-bit rgb to pixel
 *    PEXtk_FUNC_PIXEL_RGB    	  Ptrue_color  	    	 truecolor rgb to pixel
 *    PEXtk_FUNC_DITHER    	  PEXtk_default  	 dummy routine
 *    PEXtk_FUNC_TRANSPARENCY  	  PEXtk_default  	 dummy routine
 *
 *
 *  Changes from V1.0 to V1.1
 *    1. The routine names xxx_reset_zbuffer were renamed to xxx_clear_z.
 *    2. The default MBX update action was changed to Background.
 *    3. The routine PEXtk_set_update_action was added so that the user may
 *	 change the default MBX udpate action.
 *    4. A default clear_buffer routine clear_none was added.
 *    5. The PEXtk_mpex_clear_iz_all routine was removed.
 *
 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>
#include <X11/PEX5/PEXHPlib.h>

#ifdef MPEX
#include <X11/MPEX.h>
#include <X11/MPEXproto.h>
#include <X11/MPEXlib.h>
#endif

#include <X11/extensions/multibuf.h>

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"

#ifdef MPEX
	static mpxlRendererAttributes rdrNewAttribs;
#endif

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_init_buf      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes buffering routines.
 *
 *  USAGE:
 *	PEXtk_init_buf( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *	This routine sets pointers for:
 *		pointer		Extension	Available routines
 *		---------	---------	------------------------
 *		swap_buffer	MBUF		PEXtk_swap_mbuf
 *						PEXtk_copy_window
 *		create_buffer	MBUF		PEXtk_create_mbuf 
 *						PEXtk_create_back_pixmap
 *		delete_buffer	MBUF		PEXtk_delete_mbuf 
 *						PEXtk_delete_back_pixmap
 *		clear_zbuffer	MPEX		PEXtk_mpex_clear_z
 *				PEX 5.1		PEXtk_clear_z
 *				PEX 5.0		PEXtk_clear_z_5
 *		clear_buffer	MPEX		PEXtk_mpex_clear_iz
 *				PEX 5.1		PEXtk_clear_iz
 *				PEX 5.0		PEXtk_clear_iz_5
 *		        	MPEX		PEXtk_mpex_clear_i 
 *				PEX 5.1		PEXtk_clear_i
 *				PEX 5.0		PEXtk_clear_i_5
 *
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
PEXtk_init_buf(p_obj)
PEXTK_OBJECT		*p_obj;
{
    int m_op, f_ev, f_err;
    /*
     * Save the various object ptrs
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;
    p_clr	= p_obj->clr_tbl;
    p_x_data	= p_obj->x_data;

    dpy = p_x_data->dpy;
    gc  = p_x_data->gc;

#ifdef OLD
    if(XQueryExtension( dpy, "Multi-Buffering",
				&m_op, &f_ev, &f_err))
	p_support->dbl_buf = PEXtk_DBUF_MBUF;
#endif

    if( p_support->dbl_buf & PEXtk_DBUF_MBUF ){
	p__object->swap_buffer		= PEXtk_swap_mbuf;
	p__object->create_buffer	= PEXtk_create_mbuf;
	p__object->delete_buffer	= PEXtk_delete_mbuf;
    } else if( p_support->dbl_buf == PEXtk_DBUF_ES_ESCAPE ){
	p__object->swap_buffer		= PEXtk_swap_es_dblbuf;
	p__object->create_buffer	= PEXtk_create_es_dblbuf;
	p__object->delete_buffer	= PEXtk_delete_es_dblbuf;
    } else {
	p__object->swap_buffer		= PEXtk_copy_window;
	p__object->create_buffer	= PEXtk_create_back_pixmap;
	p__object->delete_buffer	= PEXtk_delete_back_pixmap;
    }

#ifdef MPEX
    p__object->clear_zbuffer		= PEXtk_mpex_clear_z;
    p__object->clear_buffer		= PEXtk_mpex_clear_iz;
    p__object->mpex_reset_state		= PEXtk_mpex_reset_state;
    p__object->push_clear_buffer	= PEXtk_mpex_clear_iz;
#else
    if( p_state->maj_ver == 5 && p_state->min_ver == 1){
	p__object->clear_zbuffer	= PEXtk_clear_z;
	p__object->clear_buffer		= PEXtk_clear_iz;
	p__object->push_clear_buffer	= PEXtk_clear_iz;
	p__object->mpex_reset_state	= PEXtk_mpex_reset_state;
    } else {
	p__object->clear_zbuffer	= PEXtk_clear_z_5;
	p__object->clear_buffer		= PEXtk_clear_iz_5;
	p__object->push_clear_buffer	= PEXtk_clear_iz;
	p__object->mpex_reset_state	= PEXtk_mpex_reset_state;
    }
#endif

    /* set default update action */
    /* UPDATE - should default clear_buffer be none with this action */
    p_state->upd_act = MultibufferUpdateActionUntouched;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+ PEXtk_create_back_pixmap +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Creates a pixmap to use as back buffer for
 *	the currently active window id.
 *
 *  USAGE:
 *	PEXtk_create_back_pixmap( w, h )
 *	int	    w		(in)	buffer width
 *	int	    h		(in)	buffer height
 *
 *  CAVEATS:
 *	This routine assumes that pstate.win[i].w_id has been
 *	set to the buffer with which the back id is bound to.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_create_back_pixmap( w, h)
int	h,w;
{
    int i = p_state->cur_w;

    p_state->win[i].back_w_id =
	XCreatePixmap( dpy, p_state->win[i].w_id,
                     w, h, p_state->win[i].depth);

    XSetForeground(dpy, gc, 
	BlackPixel(dpy,DefaultScreen(dpy)));	/* clear out the buffer */
    XFillRectangle(dpy, p_state->win[i].back_w_id, 
	gc, 0, 0, w, h);
    p_state->cur_draw = p_state->win[i].back_w_id;
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ PEXtk_delete_back_pixmap +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Deletes the back buffer pixmap 
 *	assigned to a window.
 *
 *  USAGE:
 *	PEXtk_delete_back_pixmap( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_delete_back_pixmap( )
{
    if( p_state->win[p_state->cur_w].back_w_id != 0 ){
	XFreePixmap( dpy, p_state->win[p_state->cur_w].back_w_id);
	p_state->win[p_state->cur_w].back_w_id = 0;
    }
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ PEXtk_set_buffer_update  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the MBX buffer update action.
 *
 *  USAGE:
 *	PEXtk_set_buffer_update( act )
 *	int	    act		(in)	Buffer update action to use:
 *					    PEXtk_BUFFER_UNDEFINED
 *					    PEXtk_BUFFER_BACKGROUND
 *					    PEXtk_BUFFER_UNTOUCHED
 *					    PEXtk_BUFFER_COPIED
 *
 *  CAVEATS:
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_set_buffer_update( act )
int	act;
{
    XmbufSetWindowAttributes values;

    switch( act ){
	case PEXtk_BUFFER_UNDEFINED:
		p_state->upd_act = MultibufferUpdateActionUndefined;
		break;
	case PEXtk_BUFFER_BACKGROUND:
		p_state->upd_act = MultibufferUpdateActionBackground;
		break;
	case PEXtk_BUFFER_UNTOUCHED:
		p_state->upd_act = MultibufferUpdateActionUntouched;
		break;
	case PEXtk_BUFFER_COPIED:
		p_state->upd_act = MultibufferUpdateActionCopied;
		break;
	default:
		p_state->upd_act = MultibufferUpdateActionBackground;
		break;
    }
	/* If buffers are already allocated, then change them */
    if( p_support->dbl_buf & PEXtk_DBUF_MBUF ){
	if( p_state->win[p_state->cur_w].mbuf[0] != 0 ){
	    values.update_hint = p_state->upd_act;
	    XmbufChangeWindowAttributes( dpy, p_state->win[p_state->cur_w].w_id,
		MultibufferWindowUpdateHint, &values);
	}
    }
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ PEXtk_set_clear_function +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the buffer clear function, depending on buffer types and
 *	buffer mode. This routine is called when going from SB to DB,
 * 	or from DB to SB mode. This routine effectivly saves the DB
 *	mode clear_buffer routine address when going to SB mode, and
 *	restores it when going to DB mode.
 *
 *  USAGE:
 *	PEXtk_set_clear_function( mode )
 *	int	    mode	(in)	New buffering mode
 *
 *  CAVEATS:
 *	This routine cannot be called in a sequence with the current
 *	pstate.buffer_mode equal to 'mode'.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_set_clear_function( mode )
int	mode;
{

	if( p_state->buffer_mode == mode ) return 1;

	if( mode == PEXtk_SINGLE_BUFFER ){
	    p__object->push_clear_buffer = p__object->clear_buffer;
	    p__object->clear_buffer = PEXtk_clear_area;
	} else {
	    p__object->clear_buffer = p__object->push_clear_buffer;
	}
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_create_mbuf     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Creates a Multibuffer Extension buffer
 *	to use as back buffer.
 *
 *  USAGE:
 *	PEXtk_create_mbuf( w, h )
 *	int	    w		(in)	buffer width
 *	int	    h		(in)	buffer height
 *
 *  CAVEATS:
 *	This routine assumes that pstate.win[i].w_id has been
 *	set to the buffer with which the back id is bound to.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_create_mbuf( w, h)
int	h,w;
{
    int i = p_state->cur_w;

    if(XmbufCreateBuffers( dpy, p_state->win[i].w_id, 2, 
	p_state->upd_act, MultibufferUpdateHintFrequent, 
	&p_state->win[i].mbuf[0]) != 2)
    {
	        printf(" Error... Unable to create multibuffers\n");
        	return 0;
    }
	p_state->cur_draw = p_state->win[i].back_w_id = 
	    p_state->win[i].mbuf[1];
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     PEXtk_delete_mbuf    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Deletes a Multibuffer Extension buffer  
 *	assigned to a window id.
 *
 *  USAGE:
 *	PEXtk_delete_mbuf( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
PEXtk_delete_mbuf()
{
    int	i = p_state->cur_w;
    if( p_state->win[i].mbuf[0] != 0 ){
	XmbufDestroyBuffers( dpy, p_state->win[i].w_id);
	p_state->win[i].mbuf[0] = p_state->win[i].mbuf[1] = 0;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_create_esdbl    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Uses the HP/E&S Double Buffer Escape routine
 *	to create a back buffer.
 *
 *  USAGE:
 *	PEXtk_create_esdbl( w, h )
 *	int	    w		(in)	buffer width
 *	int	    h		(in)	buffer height
 *
 *  CAVEATS:
 *	This routine assumes that pstate.win[i].w_id has been
 *	set to the buffer with which the back id is bound to.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_create_es_dblbuf( w, h)
int	h,w;
{
	int i = p_state->cur_w;
	esEscapeDblBuffer esDblBuf;

	if( p_state->win[i].w_id == 0 ) return 1;

	esDblBuf.drawable = p_state->win[i].w_id;
	esDblBuf.bufferMode = ES_RENDERER_DBLBUFFER;
	PEXEscape( dpy, ES_ESCAPE_DBLBUFFER,
		  sizeof(esEscapeDblBuffer), (char *) &esDblBuf);
	p_state->cur_draw = p_state->win[i].back_w_id = 
	    p_state->win[i].w_id;
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_delete_es_dblbuf  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. This is a null routine for the support of
 *	the ES Escape Double Buffer.
 *
 *  USAGE:
 *	PEXtk_delete_es_dblbuf( )
 *
 *  CAVEATS:
 *	This routine will set the back_w_id to zero to indicate to the swap
 *	buffer routine that the 'back buffer' has not been created yet.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_delete_es_dblbuf()
{
	int	i = p_state->cur_w;
	esEscapeDblBuffer esDblBuf;

	if( p_state->win[i].w_id == 0 ) return 1;

	esDblBuf.drawable = p_state->win[i].w_id;
	esDblBuf.bufferMode = ES_RENDERER_SINGLEBUFFER;

	PEXEscape( dpy, ES_ESCAPE_DBLBUFFER,
		  sizeof(esEscapeDblBuffer), (char *) &esDblBuf);
	p_state->win[i].back_w_id = 0;
	p_state->cur_draw = p_state->win[i].w_id;

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     PEXtk_copy_window    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Swaps buffers by copying a pixmap back 
 *	buffer onto the front window.
 *
 *  USAGE:
 *	PEXtk_copy_window( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_copy_window()
{
    XCopyArea( dpy, p_state->win[p_state->cur_w].back_w_id, 
	p_state->win[p_state->cur_w].w_id, gc,0,0,
	p_state->win[p_state->cur_w].width, 
	p_state->win[p_state->cur_w].height, 0,0);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_swap_mbuf     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Swaps buffers by calling the Multibuffer
 *	Extension routine to swap the buffers assigned to the current window.
 *
 *  USAGE:
 *	PEXtk_swap_mbuf( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_swap_mbuf()
{
  XmbufDisplayBuffers( dpy, 1, &p_state->win[p_state->cur_w].back_w_id, 0, 0);
        p_state->win[p_state->cur_w].back_w_id = 
	( p_state->win[p_state->cur_w].back_w_id == 
	  p_state->win[p_state->cur_w].mbuf[0]) ? 
		p_state->win[p_state->cur_w].mbuf[1] : 
		p_state->win[p_state->cur_w].mbuf[0];
	p_state->cur_draw = p_state->win[p_state->cur_w].back_w_id;

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   PEXtk_swap_es_dblbuf   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Swaps buffers by calling the ES Escape
 *	routine to swap the buffers assigned to the current window.
 *
 *  USAGE:
 *	PEXtk_swap_es_dblbuf( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_swap_es_dblbuf()
{
	esEscapeSwapBuffer esSwap;
	esSwap.drawable = p_state->win[p_state->cur_w].w_id;
	PEXEscape( dpy, ES_ESCAPE_SWAPBUFFER,
		  sizeof(esEscapeSwapBuffer), (char *) &esSwap);

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_clear_area      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the image plane to an rgb triplet 
 *	in single buffer mode. 
 *
 *  USAGE:
 *	PEXtk_clear_area( r, g, b )
 *	float	    r		(in)	The red value to set image buffer to
 *	float	    g		(in)	The green value to set image buffer to
 *	float	    b		(in)	The blue value to set image buffer to
 *
 *  CAVEATS:
 *	For XClearArea to work, the cur_draw must be mapped.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_clear_area( r, g, b)
float r;
float g;
float b;
{
    int x_ur, y_ur;
    int loc, width, height;
    long pix;

    PEXtk_End_Rendering( );
    loc = p_state->vpt_loc;
    x_ur = p_state->vpt[loc].left;	/* viewport x_ur */

    width  = (int) abs( p_state->vpt[loc].right - p_state->vpt[loc].left ) + 1;
    height = (int) abs( p_state->vpt[loc].top - p_state->vpt[loc].bottom ) + 1;
    y_ur = p_state->win[p_state->cur_w].height - 
		p_state->vpt[loc].top - 1; /* invert y */

    pix = (*p__object->pixel_color_rgb)( r, g, b);
    XSetForeground(dpy, gc, pix);

    if( width > p_state->win[p_state->cur_w].width ) 
        width = p_state->win[p_state->cur_w].width;
    if( height > p_state->win[p_state->cur_w].height ) 
        height = p_state->win[p_state->cur_w].height;

    XClearArea(dpy, p_state->cur_draw, x_ur, y_ur, width, height, False);
    PEXtk_Begin_Rendering( );
    /* PEXtk_reset_pex_state(1,0); */

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     PEXtk_clear_iz_5     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the image plane to an rgb triplet
 *	and initializes the z buffer. This is for PEX 5.0 servers.
 *
 *  USAGE:
 *	PEXtk_clear_iz_5( r, g, b )
 *	float	    r		(in)	The red value to set image buffer to
 *	float	    g		(in)	The green value to set image buffer to
 *	float	    b		(in)	The blue value to set image buffer to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_clear_iz_5( r, g, b)
float r;
float g;
float b;
{
    int x_ll, y_ur;
    int loc, width, height;
    long pix;

    PEXtk_End_Rendering( );
    loc = p_state->vpt_loc;
    x_ll = p_state->vpt[loc].left;	/* viewport x_ll */

    width  = (int) abs( p_state->vpt[loc].right - p_state->vpt[loc].left ) + 1;
    height = (int) abs( p_state->vpt[loc].top - p_state->vpt[loc].bottom ) + 1;
    y_ur = p_state->win[p_state->cur_w].height - 
		p_state->vpt[loc].top - 1; /* invert y */

    pix = (*p__object->pixel_color_rgb)( r, g, b);
    XSetForeground(dpy, gc, pix);
    XFillRectangle(dpy, p_state->cur_draw, gc, x_ll, y_ur, width, height);
    PEXtk_Begin_Rendering( );
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     PEXtk_clear_i_5      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the image plane to an rgb triplet.
 *	This is for PEX 5.0 servers.
 *
 *  USAGE:
 *	PEXtk_clear_i_5( r, g, b )
 *	float	    r		(in)	The red value to set image buffer to
 *	float	    g		(in)	The green value to set image buffer to
 *	float	    b		(in)	The blue value to set image buffer to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_clear_i_5( r, g, b)
float r;
float g;
float b;
{
    int x_ll, y_ur;
    int loc, width, height;
    long pix;

    PEXtk_End_Rendering( );
    loc = p_state->vpt_loc;
    x_ll = p_state->vpt[loc].left;

    width  = abs( p_state->vpt[loc].right - p_state->vpt[loc].left ) + 1;
    height = abs( p_state->vpt[loc].top - p_state->vpt[loc].bottom ) + 1;
    y_ur = p_state->win[p_state->cur_w].height - 
		p_state->vpt[loc].top - 1; /* invert y */

    pix = (*p__object->pixel_color_rgb)( r, g, b);
    XSetForeground(dpy, gc, pix);
    XFillRectangle(dpy, p_state->cur_draw, gc, x_ll, y_ur, width, height);
    /* PEXtk_reset_pex_state(1,0); */
    PEXtk_Begin_Rendering( );

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_clear_z_5     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the z buffer by doing an 
 *	End/Begin Render sequence. This is for PEX 5.0 servers.
 *
 *  USAGE:
 *	PEXtk_clear_z_5( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_clear_z_5()
{
    PEXtk_reset_pex_state(1,0);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_clear_none      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Performs no clearing of z or image planes.
 *
 *  USAGE:
 *	PEXtk_clear_none( r, g, b )
 *	float	    r		(in)	Ignored.
 *	float	    g		(in)	Ignored.
 *	float	    b		(in)	Ignored.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_clear_none( r, g, b)
float r;
float g;
float b;
{
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_clear_iz      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the image plane to an rgb triplet 
 *	and initializes the z buffer
 *
 *  USAGE:
 *	PEXtk_clear_iz( r, g, b )
 *	float	    r		(in)	The red value to set image buffer to
 *	float	    g		(in)	The green value to set image buffer to
 *	float	    b		(in)	The blue value to set image buffer to
 *
 *  DESCRIPTION:
 *	This routine will set the clear_image and clear_z flags for the
 *	renderer, do a change renderer request, and then clear the flags.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_clear_iz(r, g, b)
float r;
float g;
float b;
{
    unsigned long mask;
	PEXtk_End_Rendering( );

	p_state->rdrAttrib.clear_image = True;
	p_state->rdrAttrib.clear_z = True;
	p_state->rdrAttrib.background_color.type = PEXColorTypeRGB;
	p_state->rdrAttrib.background_color.value.rgb.red   = r;
	p_state->rdrAttrib.background_color.value.rgb.green = g;
	p_state->rdrAttrib.background_color.value.rgb.blue  = b;

	mask = PEXRAClearImage | PEXRABackgroundColor;
	if( p_state->zbuffer == PEXtk_ZBUFFER_ON )
	    mask |= PEXRAClearZ;

	PEXChangeRenderer( dpy, p_state->cur_rdr, 
			mask, &p_state->rdrAttrib);
	PEXtk_Begin_Rendering( );

	p_state->rdrAttrib.clear_image = False;
	p_state->rdrAttrib.clear_z = False;

	PEXChangeRenderer( dpy, p_state->cur_rdr, mask, 
	    &p_state->rdrAttrib);

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_clear_i       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the image plane to an rgb triplet.
 *
 *  USAGE:
 *	PEXtk_clear_i( r, g, b )
 *	float	    r		(in)	The red value to set image buffer to
 *	float	    g		(in)	The green value to set image buffer to
 *	float	    b		(in)	The blue value to set image buffer to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_clear_i( r, g, b)
float r;
float g;
float b;
{
    unsigned long mask;
	PEXtk_End_Rendering( );

	p_state->rdrAttrib.clear_image = True;
	p_state->rdrAttrib.clear_z = FALSE;
	p_state->rdrAttrib.background_color.type = PEXColorTypeRGB;
	p_state->rdrAttrib.background_color.value.rgb.red   = r;
	p_state->rdrAttrib.background_color.value.rgb.green = g;
	p_state->rdrAttrib.background_color.value.rgb.blue  = b;

	mask = PEXRAClearImage | PEXRABackgroundColor;

	PEXChangeRenderer( dpy, p_state->cur_rdr, 
			mask, &p_state->rdrAttrib);
	PEXtk_Begin_Rendering( );

	p_state->rdrAttrib.clear_image = FALSE;
	p_state->rdrAttrib.clear_z = FALSE;

	PEXChangeRenderer( dpy, p_state->cur_rdr, mask, 
	    &p_state->rdrAttrib);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_clear_z       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the z buffer by doing an 
 *	End/Begin Render sequence.
 *
 *  USAGE:
 *	PEXtk_clear_z( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_clear_z()
{
    unsigned long mask;
	PEXtk_End_Rendering( );

	p_state->rdrAttrib.clear_image = FALSE;
	p_state->rdrAttrib.clear_z = True;

	mask = PEXRAClearZ;

	PEXChangeRenderer( dpy, p_state->cur_rdr, 
			mask, &p_state->rdrAttrib);
	PEXtk_Begin_Rendering( );

	p_state->rdrAttrib.clear_image = FALSE;
	p_state->rdrAttrib.clear_z = FALSE;

	PEXChangeRenderer( dpy, p_state->cur_rdr, mask, 
	    &p_state->rdrAttrib);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_mpex_clear_z    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the z buffer using Digital's MPEX 
 *	routines.
 *
 *  USAGE:
 *	PEXtk_mpex_clear_z( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_mpex_clear_z()
{
#ifdef MPEX
    unsigned long mask;
	PEXtk_End_Rendering( );

	rdrNewAttribs.clearI = FALSE;
	rdrNewAttribs.clearZ = True;

	mask = MPEXNRAClearZ;

	MPEXChangeRenderer( dpy, p_state->cur_rdr, 
			mask, &rdrNewAttribs);
	PEXtk_Begin_Rendering( );

	rdrNewAttribs.clearI = FALSE;
	rdrNewAttribs.clearZ = FALSE;

	MPEXChangeRenderer( dpy, p_state->cur_rdr, mask, 
	    &rdrNewAttribs);
#endif
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_mpex_clear_i    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the image plane to a pixel value (pix)
 *	 using Digital's MPEX routines.
 *
 *  USAGE:
 *	PEXtk_mpex_clear_i( pix )
 *	long	    pix		(in)	The pixel value to set image buffer to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_mpex_clear_i(pix)
long pix;
{
#ifdef MPEX
    unsigned long mask;
	PEXtk_End_Rendering( );

	rdrNewAttribs.clearI = True;
	rdrNewAttribs.clearZ = FALSE;
	rdrNewAttribs.backgroundPixel = pix;

	mask = MPEXNRAClearI | MPEXNRABackgroundPixel;

	MPEXChangeRenderer( dpy, p_state->cur_rdr, 
			mask, &rdrNewAttribs);
	PEXtk_Begin_Rendering( );

	rdrNewAttribs.clearI = FALSE;
	rdrNewAttribs.clearZ = FALSE;

	MPEXChangeRenderer( dpy, p_state->cur_rdr, mask, 
	    &rdrNewAttribs);
#endif
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_mpex_clear_iz   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Clears the image plane to a pixel value (pix)
 *	 and resets the z buffer using Digital's MPEX routines.
 *
 *  USAGE:
 *	PEXtk_mpex_clear_iz( pix )
 *	long	    pix		(in)	The pixel value to set image buffer to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_mpex_clear_iz(pix)
long pix;
{
#ifdef MPEX
    unsigned long mask;
	PEXtk_End_Rendering( );

	rdrNewAttribs.clearI = True;
	rdrNewAttribs.clearZ = True;
	rdrNewAttribs.backgroundPixel = pix;

	mask = MPEXNRAClearI | MPEXNRABackgroundPixel;
	if( p_state->zbuffer == PEXtk_ZBUFFER_ON )
	    mask |= MPEXNRAClearZ;

	MPEXChangeRenderer( dpy, p_state->cur_rdr, 
			mask, &rdrNewAttribs);
	PEXtk_Begin_Rendering( );

	rdrNewAttribs.clearI = FALSE;
	rdrNewAttribs.clearZ = FALSE;

	MPEXChangeRenderer( dpy, p_state->cur_rdr, mask, 
	    &rdrNewAttribs);
#endif
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_mpex_reset_state  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module routine. Reset routiune for Digital's MPEX
 *	new renderer attributes. This routine is called in association
 *	with the PEXtk_reset_pex_state routine.
 *
 *  USAGE:
 *	PEXtk_mpex_reset_state( i, mask )
 *	int	    i		(in)	Flag set to 1 will End/Begin Rendering
 *	int	    mask	(in)	bitmask specifying which Renderer
 *					attributes are to be reset.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_mpex_reset_state(i,mask)
int	i;
int	mask;
{
#ifdef MPEX
	/*
	 * If anti-aliased lines are to be drawn, turn on aaSwitch
	 */
	if( p_state->shade_mode & PEXtk_SHADE_ANTIALIAS ){
	    rdrNewAttribs.aaSwitch = PEXOn;
	    MPEXChangeRenderer( dpy, 
		p_state->win[p_state->cur_w].Rdr_ID, 
		MPEXNRAAaSwitch, &rdrNewAttribs);
	}
	else if( p_support->allow_antialias && i ){
	    rdrNewAttribs.aaSwitch = PEXOff;
	    MPEXChangeRenderer( dpy, 
		p_state->win[p_state->cur_w].Rdr_ID, 
		MPEXNRAAaSwitch, &rdrNewAttribs);
	}
#endif
    return 0;
}

