/*
 *  $Header: /auto/nfs/sequoia/u/an/work/rws/pextk/src/pextk_db.c,v 1.1 1993/10/12 00:07:09 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                             pextk_db.c                             *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****
 *****									  *****


    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */

 /*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk Database routines.
 *
 *  Module Notes:
 *	The module controls the storing of PEXtk data. It maintains lists
 *	for:
 *		structure names 	Data is stored in a contiguous
 *					block (it is realloced when more
 *					space is needed). The PEX structure
 *					ID's are referenced by this list.
 *					The data structure also contains a
 *					pointer to the label list if any
 *					are used for that structure name.
 *					Note that name 'values' are not used
 *					as offsets into the memory block.
 *
 *		labels names		Data is stored in linked list of
 *					'data blocks' of PLABEL size. The
 *					name 'values' are not direct 
 *					offsets into the block.
 *		color table values	Data is stored in a linked list of
 *					color table blocks. The index is a
 *					direct reference, such that index 'i'
 *					is in block '(i / CTABLE_SIZE) + 1',
 *					at offset 'i % CTABLE_SIZE'
 *
 *		all primitive building data: vertex, vertex color, 
 *			vertex normal, texture vertex, polygon normal
 *
 *  Module Contents:
 *    int   DB_check_build_buffer_size
 *    int   DB_check_indicies_size
 *    int   DB_check_connectivity_size
 *    int   DB_check_build_mask		Checks vtx cnts for consistent sizes
 *    int   DB_check_struct_list	Check struct size for inserting
 *    int   DB_check_label_list		Check label list and alloc if needed
 *    int   DB_add_dls_alloc_block
 *    int   DB_add_dls_list
 *    int   DB_find_dls_name
 *    int   DB_insert_dls_name
 *    int   DB_free_struct_list		Free structure list and label list
 *    int   DB_check_bld_vert		Check vert list to add a new vert
 *    int   DB_check_bld_color		Check color list to add a new color
 *    int   DB_check_bld_normal		check normal list to add a new normal
 *    int   DB_check_bld_texture	check texture2/3 list to add a new tex
 *    int   DB_check_bld_polynorm	Check polynorm list to add a polynorm
 *    int   DB_check_index_list		Outdated
 *    int   DB_check_poly_count_list
 *    int   DB_check_bld_poly_point	Checks all bld_mask items for size
 *    int   DB_get_bld_type_string
 *    int   DB_init_bld_counts
 *    int   DB_check_dirty_mask		Check dirty_mask in immed mode.
 *    int   DB_check_all_dirty_mask	Check dirty_mask w/ rdr idle, update PC
 *    int   DB_find_struct_name
 *    int   DB_print_struct_list
 *    int   DB_find_label_name
 *    int   DB_set_struct_name		Init name to value
 *    int   DB_insert_struct_name	Insert name into table, no duplicates
 *    int   DB_delete_struct_name
 *    int   DB_set_label_name		Init label to value
 *    int   DB_clear_label_list
 *    int   DB_delete_label_list
 *    int   DB_delete_dls_name
 *    int   DB_insert_color_values
 *    int   DB_get_color_values
 *	
 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#ifdef MPEX
#include <X11/MPEX.h>
#include <X11/MPEXproto.h>
#include <X11/MPEXlib.h>
#endif

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"

/*
 *
 *                       +--------------------------+
 * ----------------------+       PEXtk_init_db      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes database routines.
 *
 *  USAGE:
 *	PEXtk_init_db( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *	This routine allocates memory for:
 *		pointer		variable	purpose
 *		---------	--------	---------------------
 *		p_state		str_list	structure id list
 *		p_build		buffer		(char *) buffer for misc data
 *		p_build		bld_vert	Pvertex list
 *				bld_normal	Pvertex_normal list
 *				bld_color	Pvertex_color list
 *				bld_texture	Pvertex_texture list
 *				bld_polynorm	Ppoly_normal list
 *		p_build		cursor_pos	Current pos for move/draw
 *
 *
 *  NOTE:
 *	The label list, p_build->str_list[name].lbl_list, is allocated
 *	when a Pmake_label routine is called. Otherwise, the list is
 *	left as NULL (empty).
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_init_db(p_obj)
PEXTK_OBJECT		*p_obj;
{
    int i;

    /*
     * Save the various object ptrs
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;
    p_clr	= p_obj->clr_tbl;
    p_x_data    = p_obj->x_data;

    p_build->bld_mask = 0;

    /* Create point/color/normal combined buffer */

    /*    n sets      3 items   item size    pt/clr/norm  */
    i = PEXtk_POLY_SIZE * 3 * sizeof(float) *     3 ;
    if(( p_build->buffer = (char *) malloc(i)) == NULL)
	return 1;

    p_build->buffer_size = i; /* size of buffer in number of bytes */

#ifdef NEW
    /* Create polylines buffer */

    if(( p_buffer->plines = (pexPolylineData *) 
	malloc(PEXtk_POLY_SIZE*sizeof(pexPolylineData))) == NULL)
	return 1;
    p_buffer->plines_size = PEXtk_POLY_SIZE;

    for(i=0;i<p_buffer->plines_size;i++)
    {
	p_buffer->plines[i].numPoints = 0;
	p_buffer->plines[i].points    = NULL;
    }

    /* Create polygons buffer */

    if(( p_buffer->ppoly = (pexPolygonData *) 
	malloc(PEXtk_POLY_SIZE*sizeof(pexPolygonData))) == NULL)
	return 1;

    p_buffer->ppoly_size = PEXtk_POLY_SIZE;
    for(i=0;i<p_buffer->ppoly_size;i++)
    {
	p_buffer->ppoly[i].numPoints = 0;
	p_buffer->ppoly[i].points    = NULL;
	p_buffer->ppoly[i].facetData = NULL;
    }
#endif

    /* Create structure id buffer */

    if((p_state->str_list = (STRUCT *)
	malloc( sizeof(STRUCT)*PEXtk_DEFAULT_STRUCT_SIZE)) == NULL)
	return 1;
    
    p_state->str_size = PEXtk_DEFAULT_STRUCT_SIZE;
    for( i=0; i< PEXtk_DEFAULT_STRUCT_SIZE; i++){
    	p_state->str_list[i].lbl_list  = (PLABEL *) NULL;
	p_state->str_list[i].cur_lbl  = (PLABEL *) NULL;
	p_state->str_list[i].last_lbl = (PLABEL *) NULL;
    }
    p_state->str_cnt  = 0;
    DB_set_struct_name( 0, PEXtk_DEFAULT_STRUCT_SIZE, -1);

#ifdef PEX4_0
    /* Create a fixed list for indexedpolygon index buffer */

    if((p_buffer->vert_index = (short *)
	malloc( sizeof(short)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_buffer->index_size = PEXtk_POLY_SIZE;

    for(i=0;i<p_buffer->index_size;i++)
	p_buffer->vert_index[i] = i;

    /* Create indexedpolygon polygon_count buffer */

    if((p_buffer->poly_Count = (short *)
	malloc( sizeof(short)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_buffer->poly_Count_size = PEXtk_POLY_SIZE;

    for(i=0;i<p_buffer->poly_Count_size;i++)
	p_buffer->poly_Count[i] = 3;
#endif

    /* Create sofas indicies buffer */

    if((p_buffer->indicies = (unsigned short *)
	malloc( sizeof(unsigned short)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_buffer->indicies_size = PEXtk_POLY_SIZE;

    /* Create sofas connectivity buffer */

    if((p_buffer->cdata = (PEXConnectivityData *)
	malloc( sizeof(PEXConnectivityData)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    /* Create sofas shorts_list buffer, same size as connectivity buffer */

    if((p_buffer->shorts_list = (PEXListOfUShort *)
	malloc( sizeof(PEXListOfUShort)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_buffer->cdata_size = PEXtk_POLY_SIZE;

    /* Initialize the connectivity buffer pointers */

    for( i = 0; i < PEXtk_POLY_SIZE; i++ ){
	p_buffer->cdata[i].count = 1;
	p_buffer->cdata[i].lists = &p_buffer->shorts_list[i];
    }

    /* Create Pbegin... vertex buffer */

    if((p_build->bld_vert = (BLD_VERT *)
	malloc( sizeof(BLD_VERT)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_build->vert_size = PEXtk_POLY_SIZE;
    p_build->vert_cnt  = 0;

    /* Create Pvertex_color... buffer */

    if((p_build->bld_color = (BLD_COLOR *)
	malloc( sizeof(BLD_COLOR)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_build->color_size = PEXtk_POLY_SIZE;
    p_build->color_cnt  = 0;

    /* Create Pvertex_normal... buffer */

    if((p_build->bld_normal = (BLD_NORMAL *)
	malloc( sizeof(BLD_NORMAL)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_build->normal_size = PEXtk_POLY_SIZE;
    p_build->normal_cnt  = 0;

    /* Create Pvertex_texture... buffer */

    if((p_build->bld_texture = (BLD_TEXTURE *)
	malloc( sizeof(BLD_TEXTURE)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_build->texture_size = PEXtk_POLY_SIZE;
    p_build->texture_cnt  = 0;

    /*
     * Create Ppoly_normal buffer. Usually only one of these will be
     * specified per polygon, however when using the Pbegin_tmesh/Pend_tmesh
     * a total of (n-2) will be specfied.
     */

    if((p_build->bld_polynorm = (BLD_NORMAL *)
	malloc( sizeof(BLD_NORMAL)*PEXtk_POLY_SIZE)) == NULL)
	return 1;

    p_build->polynorm_size = PEXtk_POLY_SIZE;
    p_build->polynorm_cnt  = 0;

    /* Init the move/draw cursor position */
    p_build->cursor_pos.x = p_build->cursor_pos.y = p_build->cursor_pos.z = 0.;

    /* Init the Ppoly_point... counters */
    p_build->pt2d_cnt   = 0;
    p_build->pt3d_cnt   = 0;
    p_build->nv_cnt     = 0;
    p_build->uv_cnt     = 0;
    p_build->nvuv_cnt   = 0;

    return 0;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_close_db      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Closes database routines.
 *
 *  USAGE:
 *	PEXtk_close_db( )
 *
 *  DESCRIPTION:
 *	This routine de-allocates memory for the items allocated in
 *	the PEXtk_init_db routine.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_close_db()
{
    if( p_build->buffer != (char *) NULL)
	free( (char *) p_build->buffer);

    p_build->buffer = NULL;
    p_build->buffer_size = 0;

    /* Free structure id buffer */

    if(p_state->str_list != (STRUCT *) NULL)
	free( (char *) p_state->str_list);
    
    p_state->str_list = NULL;
    p_state->str_size = 0;
    p_state->str_cnt  = 0;

#ifdef PEX4_0
    /* Free the fixed list for indexedpolygon index buffer */

    if(p_buffer->vert_index != (short *) NULL)
	free( (char *) p_buffer->vert_index);

    p_buffer->vert_index = NULL;
    p_buffer->index_size = 0;

    /* Free indexedpolygon polygon_count buffer */

    if(p_buffer->poly_Count != (short *) NULL)
	free( (char *) p_buffer->poly_Count);

    p_buffer->poly_Count = NULL;
    p_buffer->poly_Count_size = 0;
#endif

    /* Free sofas indicies buffer */

    if(p_buffer->indicies != (unsigned short *) NULL)
	free( (char *) p_buffer->indicies);

    p_buffer->indicies = NULL;
    p_buffer->indicies_size = 0;

    /* Free Pbegin... vertex buffer */

    if(p_build->bld_vert != (BLD_VERT *) NULL)
	free( (char *) p_build->bld_vert);

    p_build->bld_vert  = NULL;
    p_build->vert_size = 0;
    p_build->vert_cnt  = 0;

    /* Free Pvertex_color... buffer */

    if(p_build->bld_color != (BLD_COLOR *) NULL)
	free( (char *) p_build->bld_color);

    p_build->bld_color  = NULL;
    p_build->color_size = 0;
    p_build->color_cnt  = 0;

    /* Free Pvertex_normal... buffer */

    if(p_build->bld_normal != (BLD_NORMAL *) NULL)
	free( (char *) p_build->bld_normal);

    p_build->bld_normal  = NULL;
    p_build->normal_size = 0;
    p_build->normal_cnt  = 0;

    /* Free Pvertex_texture... buffer */

    if(p_build->bld_texture != (BLD_TEXTURE *) NULL)
	free( (char *) p_build->bld_texture);

    p_build->bld_texture  = NULL;
    p_build->texture_size = 0;
    p_build->texture_cnt  = 0;

    /* Free Ppoly_normal buffer */

    if(p_build->bld_polynorm != (BLD_NORMAL *) NULL)
	free( (char *) p_build->bld_polynorm);

    p_build->bld_polynorm  = NULL;
    p_build->polynorm_size = 0;
    p_build->polynorm_cnt  = 0;

    /* Clear the Ppoly_point... counters */
    p_build->pt2d_cnt   = 0;
    p_build->pt3d_cnt   = 0;
    p_build->nv_cnt     = 0;
    p_build->uv_cnt     = 0;
    p_build->nvuv_cnt   = 0;
}

/*
 *  ROUTINE:    DB_check_build_buffer_size
 *  PURPOSE:	Checks the size of the char * buffer which
 *		is used in the PEXFillArea routine.
 *  NOTE:	This routine assumes that a BLD_VERT, BLD_COLOR, and
 *		BLD_NORMAL are 3 floats each.
 *		The input value 'size' is in bytes.
 */
DB_check_build_buffer_size( size )
int	size;
{

	if( p_build->buffer_size < size ){
	    free( (char *) p_build->buffer);
	    if((p_build->buffer = (char *) malloc( size )) == NULL)
		return 1;
	    p_build->buffer_size = size;
	}
	return 0;
	    
}

/*
 *  ROUTINE:    DB_check_indicies_size
 *  PURPOSE:	Checks the size of the short * buffer which is used in
 *		the Psphere_fill routine to hold the vertex index values.
 *  SYNOPSIS	The current size of the buffer, indicies_size, is compared
 *		with the input value 'size', which is in number of items,
 *		to determine if the buffer 'indicies' needs to be free'd
 *		and malloc'd.
 */
DB_check_indicies_size( size )
int	size;
{

	if( p_buffer->indicies_size < size ){
	    free( (char *) p_buffer->indicies);
	    if((p_buffer->indicies = (unsigned short *)
		malloc( size*sizeof(short) )) == NULL)
		return 1;
	    p_buffer->indicies_size = size;
	}
	return 0;
	    
}

/*
 *  ROUTINE:    DB_check_connectivity_size
 *  PURPOSE:	Checks the size of the PEXConnectivityData buffer and the
 *		shorts_list buffer which is used in the Ppoly_index and 
 *		Psphere_fill routines to hold the vertex index values.
 *  SYNOPSIS	The current size of the buffer, cdata_size, is compared
 *		with the input value 'size', which is in number of items,
 *		to determine if the buffer 'cdata' needs to be free'd
 *		and malloc'd.
 */
DB_check_connectivity_size( size )
int	size;
{
    int i;
	if( p_buffer->cdata_size < size ){
	    if( p_buffer->cdata)
		free( (char *) p_buffer->cdata);
	    if( p_buffer->shorts_list)
		free( (char *) p_buffer->shorts_list);
	    if((p_buffer->cdata = (PEXConnectivityData *)
		malloc( size*sizeof(PEXConnectivityData) )) == NULL)
		return 1;
	    if((p_buffer->shorts_list = (PEXListOfUShort *)
		malloc( size*sizeof(PEXListOfUShort) )) == NULL)
		return 1;
	    p_buffer->cdata_size = size;
	    for( i = 0; i < size; i++ ){
		p_buffer->cdata[i].count = 1;
		p_buffer->cdata[i].lists = &p_buffer->shorts_list[i];
	    }
	}
	return 0;
	    
}

/*
 *  ROUTINE:    DB_check_build_mask
 *  PURPOSE:	Checks the build mask which is a bit encoding
 *		of calls to Pvertex (bit 0), Pvertex_color (bit 1),
 *		and Pvertex_normal (bit 2).
 *		This routine will check to make sure that the number
 *		of calls to each of these are consistent by checking
 *		the count ( xxx_cnt ) for each.
 */
/*ARGSUSED*/
int
DB_check_build_mask()
{
    switch( p_build->bld_mask ){

      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR:

	    if( p_build->color_cnt != p_build->vert_cnt ){
	      if( p_state->debug_mode )
	        fprintf(stderr,
	         "ERROR... Pvertex count not equal to Pvertex_color count\n");
	        return 1;
	    }
	  break;
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_NORM:

	    if( p_build->normal_cnt != p_build->vert_cnt ){
	      if( p_state->debug_mode )
	        fprintf(stderr,
	         "ERROR... Pvertex count not equal to Pvertex_normal count\n");
	        return 1;
	    }
	  break;
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR|PEXtk_BLD_SET_NORM:
	    if( p_build->color_cnt != p_build->vert_cnt &&
		p_build->normal_cnt != p_build->vert_cnt){
	      if( p_state->debug_mode )
	        fprintf(stderr,
	        "ERROR... Pvertex count not equal to Pvertex_color/Pvertex_normal count\n");
	        return 1;
	    }
	  break;
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR|PEXtk_BLD_SET_NORM|
	   PEXtk_BLD_SET_TEXTURE:
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR|PEXtk_BLD_SET_NORM|
	   PEXtk_BLD_SET_TEXTURE3:
	    if( p_build->color_cnt != p_build->vert_cnt &&
		p_build->normal_cnt != p_build->vert_cnt &&
		p_build->texture_cnt != p_build->vert_cnt){
	      if( p_state->debug_mode )
	        fprintf(stderr,
	        "ERROR... Pvertex count not equal to Pvertex_color/Pvertex_normal/Pvertex_texture count\n");
	        return 1;
	    }
	  break;
	default:
		/* illegal combination */
	  break;
    }
    return 0;
}

/*
 *  ROUTINE:    DB_check_struct_list
 *  PURPOSE:	This routine checks the structure list to see if another
 *		structure id can be added to the list. If the list is
 *		not large enough, then the list is reallocated to the
 *		the current size + the default block size.
 */
DB_check_struct_list()		/* routine to check struct size for inserting */
{				/* a new structure id in list */
    if(p_state->str_cnt+1 > p_state->str_size)
    {
	if((p_state->str_list = (STRUCT *)realloc( (char *)p_state->str_list,
	  sizeof(STRUCT)*(p_state->str_size+PEXtk_DEFAULT_STRUCT_SIZE)))==NULL)
		return 1;
        DB_set_struct_name(p_state->str_size,PEXtk_DEFAULT_STRUCT_SIZE,-1);
        p_state->str_size += PEXtk_DEFAULT_STRUCT_SIZE;
    }
    return 0;
}

/*
 *  ROUTINE:    DB_check_label_list
 *  PURPOSE:	This routine checks the label list to see if another
 *		label can be added to structure 'i'. If the lbl_list 
 *		is NULL, then a block is allocated. If the current block
 *		is full, then a new block is allocated and the last
 *		label list pointer (last_lbl) is set to point to the newly
 *		allocated block.
 */
int
DB_check_label_list(i)		/* Check label list size, alloc if needed */
int	i;
{
    PLABEL *ptr;
    if( p_state->str_list[i].lbl_list == (PLABEL *) NULL){
      if((ptr = (PLABEL *) malloc( sizeof(PLABEL))) == (PLABEL *) NULL)
	return 1;
    
	ptr->cnt  			    = 0;
	ptr->next  			    = (PLABEL *) NULL;
	p_state->str_list[i].lbl_list 	    = ptr;
	p_state->str_list[i].last_lbl	    = ptr;
	p_state->str_list[i].lbl_blocks     = 1;
	p_state->str_list[i].lbl_size       = LABEL_SIZE;
    } else
	/* check if this block has any more room */
    if( p_state->str_list[i].last_lbl->cnt+1 > 
	LABEL_SIZE )
    {
	if((ptr = (PLABEL *)malloc( sizeof(PLABEL)))==(PLABEL *) NULL)
		return 1;
	ptr->cnt  = 0;
	ptr->next = (PLABEL *) NULL;
	p_state->str_list[i].last_lbl->next	= ptr;
	p_state->str_list[i].last_lbl		= ptr;
        DB_set_label_name(ptr,LABEL_SIZE,PEXtk_STR_EMPTY_LABEL);
        p_state->str_list[i].lbl_blocks += 1;
        p_state->str_list[i].lbl_size += LABEL_SIZE;
    }
    return 0;

}

/*
 *  ROUTINE:    DB_add_dls_alloc_block
 *  PURPOSE:	This routine puts a pointer to a newly allocated dpy
 *		list node block into the alloc_node ptr list.
 *  DESCRIPTION:
 *		This routine first checks to make sure that there is enough
 *		room in the DLS_ALLOC_NODE block. To do this, the node_ptr
 *		is checked if it has been allocated yet. If the node_ptr 
 *		is NULL (i.e., it has not been allocated yet), then a
 *		block is allocated. If the current block is full,
 *		then a new block is allocated and the current dls block
 *		list pointer (cur_dpy_list_block) is set to point to the 
 *		newly allocated block.
 */
int
DB_add_dls_alloc_block( nnode )
DPY_LIST_NODE	*nnode;			/* new node to add to list */
{
    DLS_ALLOC_NODE *ptr;
    int i;
	/* has the block been allocated yet ? */
    if(p_dl->node_ptr == NULL ){
      if((ptr = (DLS_ALLOC_NODE *) malloc( sizeof(DLS_ALLOC_NODE))) == 
			(DLS_ALLOC_NODE *) NULL)
	return 1;
    
	ptr->cnt		= 0;
	ptr->next		= (DLS_ALLOC_NODE *) NULL;
	p_dl->node_ptr		= ptr;
	p_dl->node_block_cnt	= 1;
	p_dl->cur_node_ptr     	= ptr;
    } else
	/* check if this block has any more room */
    if(p_dl->cur_node_ptr->cnt+1 > DLS_SIZE)
    {
	if((ptr = (DLS_ALLOC_NODE *)malloc( sizeof(DLS_ALLOC_NODE))) ==
		(DLS_ALLOC_NODE *) NULL)
	    return 1;
	ptr->cnt  		  = 0;
	ptr->next 		  = (DLS_ALLOC_NODE *) NULL;
	p_dl->cur_node_ptr->next  = ptr;
	p_dl->cur_node_ptr     	  = ptr;
        /* DB_set_dls_name(ptr,DLS_SIZE,-1); */
        p_dl->node_block_cnt++;		/* incr. block count */
    }

    i = p_dl->cur_node_ptr->cnt;
    p_dl->cur_node_ptr->node[i] = nnode;
    p_dl->cur_node_ptr->cnt++;

    return 0;

}

/*
 *  ROUTINE:    DB_add_dls_list
 *  PURPOSE:	This routine is used by Pcreate_dpy_list to verify whether
 *		a display list may be opened.
 *		This routine will check the dls name list block for a
 *		display list with a name of 'name'. If it does exist,
 *		then an error is returned.
 *		points to a valid name_block. To do this, the name_block
 *		is checked if it has been allocated yet. If the name_block 
 *		is NULL (i.e., it has not been allocated yet), then a
 *		block is allocated. If the current block is full,
 *		then a new block is allocated and the current name block
 *		list pointer (cur_name_list_block) is set to point to the 
 *		newly allocated block.
 */
int
DB_add_dls_list( name )
int	name;
{
    DLS_NAME *block_ptr;

    if(DB_find_dls_name( name, &block_ptr ) == -1 ) /* error if name exists*/
	return 1;
    DB_insert_dls_name(name);

	return 0;

}

/*
 *  ROUTINE:    DB_find_dls_name
 *  PURPOSE:	Returns -1 if name does not exist, index if it does exist
 */
int
DB_find_dls_name( name, block_ptr )
int	name;
DLS_NAME **block_ptr;
{

    int i;
    DLS_NAME *ptr;

    ptr = p_dl->name_block;

        do {
	 for(i=0;i<ptr->cnt;i++)
	    if( ptr->name[i] == name ){
		*block_ptr = ptr;
		return (i);
	    }
	    if( ptr->name[i] == PEXtk_DLS_EMPTY_NAME )
		return (-1);
	} while ((ptr = ptr->next) != NULL);

    return -1;
}

/*
 *  ROUTINE:    DB_insert_dls_name
 *  PURPOSE:	This routine will insert 'name' into the name block.
 *		It will also make sure that cur_name points to a
 *		valid name_block. To do this, the name_block is
 *		checked if it has been allocated yet. If the name_block 
 *		is NULL (i.e., it has not been allocated yet), then a
 *		block is allocated. If the current block is full,
 *		then a new block is allocated and the current name block
 *		list pointer (cur_name) is set to point to the 
 *		newly allocated block.
 */
int
DB_insert_dls_name(name)
int	name;
{
    DLS_NAME *ptr;
    int i;
	/* has the block been allocated yet ? */
    if(p_dl->name_block == NULL ){
      if((ptr = (DLS_NAME *) malloc( sizeof(DLS_NAME))) == 
			(DLS_NAME *) NULL)
	return 1;
    
	ptr->next		= (DLS_NAME *) NULL;
	p_dl->name_block	= ptr;
	p_dl->cur_name_ptr     	= ptr;
	p_dl->name_block_cnt	= 1;
	ptr->cnt		= 0;
    } else
	/* check if this block has any more room */
    if(p_dl->cur_name_ptr->cnt+1 > DLS_SIZE)
    {
	if((ptr = (DLS_NAME *)malloc( sizeof(DLS_NAME))) ==
		(DLS_NAME *) NULL)
	    return 1;
	ptr->next 		 = (DLS_NAME *) NULL;
	p_dl->cur_name_ptr->next = ptr;
	p_dl->cur_name_ptr     	 = ptr;
        p_dl->name_block_cnt++;		/* incr. block count */
	ptr->cnt		 = 0;
        /* DB_set_name(ptr,DLS_SIZE,-1); */
    }

    i = p_dl->cur_name_ptr->cnt;
    p_dl->cur_name_ptr->name[i] = name;
    p_dl->cur_name_ptr->cnt++;
    DLS_create_dpy_list( name );

    return 0;

}

/*
 *  ROUTINE:	DB_free_struct_list
 *  PURPOSE:	This routine deallocates all label list memory associated
 *		with all structures (i.e., the lbl_list pointer in each
 *		str_list[name] is traversed and free'd.
 *
 */
DB_free_struct_list()	/* routine to free structure list and label list */
{
    int i;
    PLABEL *ptr, *ptr_n;

    for(i=0;i<p_state->str_size;i++){
	if( (ptr=p_state->str_list[i].lbl_list) != NULL){
	    do {
		ptr_n = ptr->next;
		free( (char *) ptr);
		ptr = ptr_n;
		} while (ptr_n != NULL);
	    p_state->str_list[i].lbl_list = NULL;
	}
    }
    free( (char *) p_state->str_list);
    p_state->str_list = NULL;
}

DB_check_bld_vert(n)	/* routine to check vert list to add a new vert */
int n;			/* number of items to add */
{
    int new_size;
    if(p_build->vert_cnt+n > p_build->vert_size)
    {
	if( p_build->vert_cnt+n > p_build->vert_size+PEXtk_POLY_SIZE)
	    new_size = p_build->vert_cnt + n;
	else
	    new_size = p_build->vert_size + PEXtk_POLY_SIZE;
	if((p_build->bld_vert = (BLD_VERT *)realloc( (char*)p_build->bld_vert,
	  sizeof(BLD_VERT)*(new_size)))==NULL)
		return 1;
        p_build->vert_size = new_size;
    }
    return 0;
}

DB_check_bld_color(n)	/* routine to check color list to add 'n' new colors */
int n;			/* number of items to add */
{
    int new_size;
    if(p_build->color_cnt+n > p_build->color_size)
    {
	if( p_build->color_cnt+n > p_build->color_size+PEXtk_POLY_SIZE)
	    new_size = p_build->color_cnt + n;
	else
	    new_size = p_build->color_size + PEXtk_POLY_SIZE;
	if((p_build->bld_color = (BLD_COLOR*)realloc((char*)p_build->bld_color,
	  sizeof(BLD_COLOR)*(new_size)))==NULL)
		return 1;
        p_build->color_size = new_size;
    }
    return 0;
}

DB_check_bld_normal(n)  /* routine to check normal list to add a new normal */
int	n;		/* number of items to add */
{
    int new_size;
    if(p_build->normal_cnt+n > p_build->normal_size)
    {
	if( p_build->normal_cnt+n > p_build->normal_size+PEXtk_POLY_SIZE)
	    new_size = p_build->normal_cnt + n;
	else
	    new_size = p_build->normal_size + PEXtk_POLY_SIZE;
	if((p_build->bld_normal = (BLD_NORMAL *) realloc( (char *)
		p_build->bld_normal, sizeof(BLD_NORMAL)*(new_size)))==NULL)
		return 1;
        p_build->normal_size = new_size;
    }
    return 0;
}

DB_check_bld_texture(n) /* routine to check texture2/3 list to add a new tex */
int	n;		/* number of items to add */
{
    int new_size;
    if(p_build->texture_cnt+n > p_build->texture_size)
    {
	if( p_build->texture_cnt+n > p_build->texture_size+PEXtk_POLY_SIZE)
	    new_size = p_build->texture_cnt + n;
	else
	    new_size = p_build->texture_size + PEXtk_POLY_SIZE;
	if((p_build->bld_texture = (BLD_TEXTURE *)realloc( (char *)
		p_build->bld_texture, sizeof(BLD_TEXTURE)*(new_size)))==NULL)
		return 1;
        p_build->texture_size = new_size;
    }
    return 0;
}

DB_check_bld_polynorm()	/* routine to check polynorm list to add a polynorm */
{
    if(p_build->polynorm_cnt+1 > p_build->polynorm_size)
    {
	if((p_build->bld_polynorm = (BLD_NORMAL *)realloc( (char *)
	  p_build->bld_polynorm,
	  sizeof(BLD_NORMAL)*(p_build->polynorm_size+PEXtk_POLY_SIZE)))==NULL)
		return 1;
        p_build->polynorm_size += PEXtk_POLY_SIZE;
    }
    return 0;
}

/*
 *  ROUTINE:	DB_check_index_list
 *  PURPOSE:	This routine will check the vert_index buffer to see if it
 *		can handle 'size' more items. If it cannot, it will free
 *		the memory, and malloc 'size' more items. The vert_index
 *		buffer will not be initialized.
 */
DB_check_index_list(size)
int	size;		/* number of items to put in list */
{
#ifdef PEX4_0
    if( p_buffer->index_size >= size )
	return;

    free( (char *) p_buffer->vert_index);

    if((p_buffer->vert_index = (short *)
	malloc( sizeof(short)*size)) == NULL)
	return 1;

    p_buffer->index_size = size;
    /* for(i=0;i<p_buffer->index_size;i++) */
	/* p_buffer->vert_index[i] = i; */
#endif
}

/*
 *  ROUTINE:	DB_check_poly_count_list
 *  PURPOSE:	This routine checks the poly_Count buffer to 
 *		ensure that it can hold 'size' number of items.
 *  NOTE:	This routine is used for Psphere_fill
 */
DB_check_poly_count_list(size)
int	size;		/* number of items to put in list */
{
#ifdef PEX4_0
    int i;

    if( p_buffer->poly_Count_size >= size )
	return;

    free( (char *) p_buffer->poly_Count);

    if((p_buffer->poly_Count = (short *)
	malloc( sizeof(short)*size)) == NULL)
	return 1;

    p_buffer->poly_Count_size = size;
#endif
}

/*
 *  ROUTINE:	DB_check_bld_poly_point
 *  PURPOSE:	This routine checks the Ppoly_point build mask to
 *		ensure that all vertex lists are large enough.
 *  RETURNS:	0 if successful, 1 if memory error
 */
int
DB_check_bld_poly_point()
{
    int i;
	switch ( p_build->bld_mask ){
          case PEXtk_BLD_PNT_2D:
          case PEXtk_BLD_PNT_3D:
		return(DB_check_bld_vert(1));
		/* break; */
          case PEXtk_BLD_PNT_NV:
		i = DB_check_bld_vert(1);
		if( i ) return i;
		return (DB_check_bld_normal(1));
		/* break; */
          case PEXtk_BLD_PNT_UV:
		i = DB_check_bld_vert(1);
		if( i ) return i;
		return (DB_check_bld_texture(1));
		/* break; */
          case PEXtk_BLD_PNT_NV_UV:
		i = DB_check_bld_vert(1);
		if( i ) return i;
		i = DB_check_bld_normal(1);
		if( i ) return i;
		return (DB_check_bld_texture(1));
		/* break; */
	  default:
		return (1);
	}
}

char *
DB_get_bld_type_string()
{

	switch ( p_build->bld_mask ){
          case PEXtk_BLD_SET_VERT:
		return ("Pvertex");
          case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR:
		return ("Pvertex:Pvertex_color");
          case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR|PEXtk_BLD_SET_NORM:
		return ("Pvertex:Pvertex_color:Pvertex_normal");
          case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR|PEXtk_BLD_SET_NORM|
	     PEXtk_BLD_SET_TEXTURE:
		return ("Pvertex:Pvertex_color:Pvertex_normal:Pvertex_texture");
          case PEXtk_BLD_PNT_2D:
		return ("Ppoly_point_2d");
          case PEXtk_BLD_PNT_2D|PEXtk_BLD_SET_COLOR:
		return ("Ppoly_point_2d/color");
          case PEXtk_BLD_PNT_3D:
		return ("Ppoly_point_3d");
          case PEXtk_BLD_PNT_3D|PEXtk_BLD_SET_COLOR:
		return ("Ppoly_point_3d/color");
          case PEXtk_BLD_PNT_NV:
		return ("Ppoly_point_nv");
          case PEXtk_BLD_PNT_NV|PEXtk_BLD_SET_COLOR:
		return ("Ppoly_point_nv/color");
          case PEXtk_BLD_PNT_UV:
		return ("Ppoly_point_uv");
          case PEXtk_BLD_PNT_UV|PEXtk_BLD_SET_COLOR:
		return ("Ppoly_point_uv/color");
          case PEXtk_BLD_PNT_NV_UV:
		return ("Ppoly_point_nv_uv");
          case PEXtk_BLD_PNT_NV_UV|PEXtk_BLD_SET_COLOR:
		return ("Ppoly_point_nv_uv/color");
	  default:
		return ("unknown");
	}
}

/*
 *  ROUTINE:	DB_init_bld_counts
 *  PURPOSE:	This routine initializes the counts and build masks for
 *		the Pbegin_poly/Pend_poly and Ppoly_point_xxx/Ppoly_close
 *		sequences.
 */
int
DB_init_bld_counts()
{
	p_build->bld_mode   = PEXtk_BLD_NONE;
	p_build->bld_mask   = 0;
	p_build->vert_cnt   = 0;
	p_build->color_cnt  = 0;
	p_build->normal_cnt = 0;
	p_build->texture_cnt = 0;
	p_build->pt2d_cnt   = 0;
	p_build->pt3d_cnt   = 0;
	p_build->uv_cnt     = 0;
	p_build->nv_cnt     = 0;
	p_build->nvuv_cnt   = 0;
	p_build->polynorm_cnt = 0;

	return 0;
}

/*
 *  ROUTINE:	DB_check_dirty_mask
 *  PURPOSE:	This routine checks the dirty mask
 *  SYNOPSIS:   The state dirty mask, p_state->dirty_mask is checked
 *		against the input mask to determine if the dependant
 *		PEX command needs to be invoked.
 *  CAVEATS:	This routine is only effective in immediate mode.
 */
int
DB_check_dirty_mask( mask )
int	mask;	/* current mask to check against */
{
    if( ! IMMEDIATE_MODE( p_state->sys_mode )) return 1;
    /* for( i=0; i < PEXtk_DIRTY_N_BITS; i++, i_mask <<= 1 ) */
      switch( mask ){
	case PEXtk_DIRTY_VPT:
		break;
	case PEXtk_DIRTY_MTX:
		PEXtk_Update_Matrix(1);
		p_state->dirty_mask &= ~PEXtk_DIRTY_MTX;
		break;
	case PEXtk_DIRTY_GEOM:
		break;
	case PEXtk_DIRTY_COLOR:
		break;
	case PEXtk_DIRTY_VIEW:
		break;
	case PEXtk_DIRTY_REFL_AMB:
		break;
	case PEXtk_DIRTY_REFL_DIF:
		break;
	case PEXtk_DIRTY_REFL_SPEC:
		break;
	case PEXtk_DIRTY_REFL_EXP:
		break;
	case PEXtk_DIRTY_MRKR_TYPE:
		break;
    }
    return 0;
}

/*
 *  ROUTINE:	DB_check_all_dirty_mask
 *  PURPOSE:	This routine checks and updates the dirty mask
 *  SYNOPSIS:   The state dirty mask, p_state->dirty_mask is checked
 *		for all active bits to determine if the dependant
 *		Pipeline Context needs to be updated.
 *  CAVEATS:	This routine is effective when the renderer is idle.
 */
int
DB_check_all_dirty_mask( )
{
	if( p_state->dirty_mask & PEXtk_DIRTY_MTX){
		PEXtk_Update_Matrix(0);
		p_state->dirty_mask &= ~PEXtk_DIRTY_MTX;
	}
    return 0;
}

/*
 *  ROUTINE: DB_find_struct_name
 *  PURPOSE: Returns the index in the structure list for structure 'name'
 *  NOTE:    Since the structure list is an ascending sorted list, a more
 *	     efficient search method could be used, such as a binary search.
 */
int
DB_find_struct_name(name)
int	name;
{
	int i,delta,low,high;

	/*
	 * check if 'name' could also be the 'index' where it is
	 */
	if( name < p_state->str_cnt && name  >= 0 ){
	    if( p_state->str_list[name].name == name ){
		/* printf("Found name %d at index %d\n", name,name); */
		return name;
	    }
	    	/* check if things started at one instead */
		i = name - 1;
	    if( i < p_state->str_cnt && name > 0 ){
	        if( p_state->str_list[i].name == name ){
		    /* printf("Found name %d at index %d\n", name,name-1); */
		    return (i);
		}
	    }
	}
#ifdef USE_OLD_SEARCH
	for(i=0;i<p_state->str_cnt;i++){
	    if( p_state->str_list[i].name == name )
		return (i);
	    else if( p_state->str_list[i].name > name )
		return -1;
	}
#else
	i = p_state->str_cnt/2;
	if( p_state->str_list[i].name == name )
		return (i);
	high = p_state->str_cnt;
	low = 0;
	while(low != high){
	    if( p_state->str_list[i].name > name ){
		delta = (i-low)/2;
		if( delta == 0 )
		    if( p_state->str_list[low].name == name )
			return low;
		    else
			return -1;
		high = i;
		i = i - delta;
	    } else {
		delta = (high-i)/2;
		if( delta == 0 )
		    if( p_state->str_list[high].name == name )
			return high;
		    else
			return -1;
		low = i;
		i = i + delta;
	    }
	    if( p_state->str_list[i].name == name )
		return (i);
	}
#endif

	return -1;
}

/*
 *  ROUTINE: DB_print_struct_list
 */
int
DB_print_struct_list()
{
	int i;

	for(i=0;i<p_state->str_cnt;i++)
	    printf(" name %d, id %x\n", p_state->str_list[i].name,
	        p_state->str_list[i].id);

	return 0;
}

/*
 *  ROUTINE: DB_find_label_name
 *  PURPOSE: Finds the label 'label' in the currently opened structure.
 *	     If the label is found, then the struct_list[cur_str].cur_lbl
 *	     is set to point to the label block of the label. The index
 *	     into that block is returned.
 *  RETURNS: The index into the cur_lbl block of the label.
 */
int
DB_find_label_name(name,label)
int	name;
int	label;
{
	int i;
	PLABEL *ptr;

	i = DB_find_struct_name(name);
	if( i == -1 )
	    return -1;

	ptr = p_state->str_list[i].lbl_list;
	if( ptr == NULL ) return -1;

        do {				/* traverse the label block list */
	 p_state->str_list[i].cur_lbl = ptr;
	 for(i=0;i<ptr->cnt;i++)
	    if( ptr->label[i] == label )
		return (i);
	    /* if( ptr->label[i] == PEXtk_STR_EMPTY_LABEL ) */
		/* return (-1); */
	} while ((ptr = ptr->next) != NULL);

	return -1;
}

/*
 *  ROUTINE: DB_add_label_list
 *  PURPOSE: Adds the label 'label' in the currently opened structure.
 *	     The struct_list[cur_str].cur_lbl is set to point to the
 *	     label block of the label. 
 *  RETURNS: 0 if successful, 1 if memory error, 2 if last_lbl not set
 */
int
DB_add_label_list(loc,label)
int	loc;
int	label;
{
    PLABEL *ptr;
            /* make sure there is room in the label list */
        if(DB_check_label_list(loc))
            return 1; /* memory alloc error */
        if( p_state->str_list[loc].last_lbl != NULL ){
            ptr = p_state->str_list[loc].last_lbl;
            ptr->label[ptr->cnt] = label;
            ptr->attr[ptr->cnt] = p_state->attr->i_style_str;
            ptr->attr[ptr->cnt] |= (p_state->attr->edgeFlag_str << 4);
            ptr->attr[ptr->cnt] |= (p_state->attr->interp_meth_str << 8);
            ptr->cnt++;
            p_state->str_list[loc].cur_lbl = ptr;
	    return 0;
        } else return 2;
}

int
DB_print_label_list(loc,label)
int	loc;
int	label;
{
    int i;
    PLABEL *ptr;
    printf("Label list for struct loc %d, loc %d\n",
	p_state->str_list[loc].name,loc);

        ptr = p_state->str_list[loc].lbl_list;
	if( ptr == (PLABEL *) NULL ){
	    printf("print_label list for label %d is NULL\n",label);
	    return 1;
	} else
	    printf("print_label list label %d has %d items\n",label,ptr->cnt);
        for(i=0; i< ptr->cnt; i++)
            printf(" lbl %d, style %d\n",ptr->label[i],ptr->attr[i]);

	return 0;
}

/*
 *  ROUTINE: DB_set_struct_name
 *  PURPOSE: Sets a range of a structure list to a value 
 */
int
DB_set_struct_name(cur,size,value)/* init name to value */
int	cur;
int	size;
int	value;
{
    int i;

    for(i=cur; i < size; i++)
	p_state->str_list[i].name = value;

    return 0;
}

/*
 *  ROUTINE: DB_insert_struct_name
 *  PURPOSE: Inserts a structure name into the list by first searching for
 *	     the name in the list that is greater than 'name', and then
 *	     moving everyone down in the list.
 *  NOTE:    It is mandatory that the routine DB_check_struct_list be called
 *	     BEFORE this routine to determine if there is enough space
 *	     in the list to add a structure name.
 */
int
DB_insert_struct_name(name)/* insert name into table, no duplicates*/
int	name;
{
	int i, j;
	for(i=0;i<p_state->str_cnt;i++)
	    if( p_state->str_list[i].name > name ){
		for(j=p_state->str_cnt;j>i;j-- ){ /* move everyone down */
		  memcpy( (char *) &p_state->str_list[j], 
			(char *) &p_state->str_list[j-1],
			sizeof(STRUCT));
		}
		p_state->str_cnt++;
		p_state->str_list[i].name = name;
		p_state->str_list[i].empty = 0;	  /* def. is not empty */
		p_state->str_list[i].attr = p_state->attr->i_style_str;
		p_state->str_list[i].attr |=(p_state->attr->edgeFlag_str<<4);
		p_state->str_list[i].attr |=(p_state->attr->interp_meth_str<<8);
		return i;
	    } else
	      if( p_state->str_list[i].name == -1 ){ /* open spot */
		p_state->str_cnt++;
		p_state->str_list[i].name = name;
		p_state->str_list[i].empty = 0;	  /* default is not empty */
		p_state->str_list[i].attr = p_state->attr->i_style_str;
		p_state->str_list[i].attr |=(p_state->attr->edgeFlag_str<<4);
		p_state->str_list[i].attr |=(p_state->attr->interp_meth_str<<8);
		return i;
	      }

	/* end of items in list */
	p_state->str_list[p_state->str_cnt].name = name;
	p_state->str_list[p_state->str_cnt].empty = 0;/* def. is not empty */
	p_state->str_list[i].attr = p_state->attr->i_style_str;
	p_state->str_list[i].attr |= (p_state->attr->edgeFlag_str << 4);
	p_state->str_list[i].attr |= (p_state->attr->interp_meth_str << 8);
	p_state->str_cnt++;
	return (p_state->str_cnt-1);

}

/*
 *  ROUTINE: DB_delete_struct_name
 *  PURPOSE: Deletes a structure name from the list by first finding it
 *	     (it is not required that the calling routine first find out
 *	     if it exists), and then moving everyone down in the ordered
 *	     list. Note that the structure id is deleted also.
 *
 *  RETURNS: index of structure removed ( >= 0), -1 if error
 */
int
DB_delete_struct_name(name)
int	name;
{
	int i;
	for(i=0;i<p_state->str_cnt;i++){
	    if( p_state->str_list[i].name == name ){
		if( p_state->str_list[i].id != 0 )
		    PEXDestroyStructures( p__object->x_data->dpy, 
			1, &p_state->str_list[i].id);
		DB_delete_label_list(i); /* free the label list */
		for(;i<p_state->str_cnt-1;i++ ){ /* move everyone down */
		  memcpy( (char *) &p_state->str_list[i],
			(char *) &p_state->str_list[i+1],
			sizeof(STRUCT));
		}
		p_state->str_cnt--;
			/* set last item to empty name */
		p_state->str_list[i].name = PEXtk_STR_EMPTY_NAME; 
		p_state->str_list[i].cur_lbl = p_state->str_list[i].lbl_list =
		  p_state->str_list[i].last_lbl = NULL;	/* clear lbl list ptr*/
		return i;
	    } else
	      if( p_state->str_list[i].name > name )/* have we gone past it ? */
		return -1;
	}

	return -1;

}

/*
 *  ROUTINE: DB_set_label_name
 *  PURPOSE: Sets a range of a label list to a value. This routine is
 *	     used to initialze the label list. It will also init the style.
 */
int
DB_set_label_name(ptr,size,value)/* init label to value */
PLABEL	*ptr;
int	size;
int	value;
{
    int i;

    for(i=0; i < size; i++){
	ptr->label[i] = value;
	ptr->attr[i] = 0;
    }
    return 0;
}

/*
 *  ROUTINE: DB_clear_label_list
 *  PURPOSE: Clears an entire label list for a structure by init'ing values 
 *	     to PEXtk_STR_EMPTY_LABEL. It also resets the style to 0.
 */
int
DB_clear_label_list(i)
int	i;
{
	int j;
	PLABEL *ptr;

	ptr = p_state->str_list[i].lbl_list;
	if( ptr == NULL ) return 1;

	do {
	  for(j=0;j<ptr->cnt;j++){
	    ptr->label[j] = PEXtk_STR_EMPTY_LABEL;
	    ptr->attr[j] = 0;
	  }
	  ptr->cnt = 0;
	  p_state->str_list[i].last_lbl = ptr; /* set last label ptr */
	} while ((ptr = ptr->next) != NULL);

	/* reset current label list to first label list */
	p_state->str_list[i].cur_lbl = p_state->str_list[i].lbl_list;

    return 0;
}

/*
 *  ROUTINE: DB_delete_label_list
 *  PURPOSE: Deletes an entire label list for a structure by freeing the 
 *	     label list pointers.
 */
int
DB_delete_label_list(i)
int	i;
{
	PLABEL *ptr, *fptr;

	ptr = p_state->str_list[i].lbl_list;
	if( ptr == (PLABEL *) NULL ) return 0;

	do {
	  fptr = ptr;
	  ptr = ptr->next;
	  free( (char *) fptr);
	} while (ptr != NULL);

    return 0;
}

/*
 *  ROUTINE: DB_delete_dls_name
 *  PURPOSE: Deletes a dpy list name by removing it from the name block 
 *	     list, and taking the last item in the list and moving it to
 *	     its location, and init'ing the last items location, and 
 *	     decrementing the last block 'cnt'.
 *  NOTE:    The routine DB_find_dls_name must be called previous to this
 *	     routine so that the block_ptr (i.e., the block in which the
 *	     dls 'name' name is, and the index 'i' may be passed in.
 */
int
DB_delete_dls_name( i, name, block_ptr)
int	i;
DLS_NAME *block_ptr;
int	name;
{
	int lname;
	DLS_NAME *ptr;

	ptr = p_dl->name_block;
	if( ptr == NULL ) return 1;

	while ( ptr->next != NULL)	/* look for last block */
	    ptr=ptr->next;

	lname			= ptr->name[ptr->cnt];
	ptr->name[ptr->cnt]	= PEXtk_DLS_EMPTY_NAME;
	ptr->cnt--;
	block_ptr->name[i] = lname;

    return 0;
}

/*
 *  ROUTINE:    DB_insert_color_values
 *  PURPOSE:	This routine will insert 'col' into the color table block.
 *		The color table is arranged diffently from the other
 *		tables in that it is not a 'sparse' or 'packed' table.
 *		The 'index' will place the 'col' at the index into the
 *		array.
 *		Blocks are allocated until the 'index' value is obtained
 *		(using num_blocks * CTABLE_SIZE ).
 *		Indexing begins at 0.
 */
int
DB_insert_color_values( index, col)
int	index;
PCOLOR	*col;
{
    CTABLE_LIST *ptr, *c_ptr;
    int i;
	/* has the block been allocated yet ? */
    if(p_clr->clr_block == NULL ){
      if((ptr = (CTABLE_LIST *) malloc( sizeof(CTABLE_LIST))) == 
			(CTABLE_LIST *) NULL)
	return 1;
    
	ptr->next		= (CTABLE_LIST *) NULL;
	p_clr->clr_block	= ptr;
	p_clr->cur_block     	= ptr;
	p_clr->block_cnt	= 1;
	ptr->cnt		= 0;
    }

	/* allocate blocks until at the block to contain 'index' */
    c_ptr = NULL;
    while ( index >= (p_clr->block_cnt * CTABLE_SIZE) )
    {
	if((ptr = (CTABLE_LIST *)malloc( sizeof(CTABLE_LIST))) ==
		(CTABLE_LIST *) NULL)
	    return 1;
	ptr->next 		 = (CTABLE_LIST *) NULL;
	p_clr->cur_block->next   = ptr;
	p_clr->cur_block	 = ptr;
	/* use c_ptr to indicate current block known due to it was allocated */
	c_ptr			 = ptr;
        p_clr->block_cnt++;	/* incr. block count */
	ptr->cnt		 = 0;
    }

    if( c_ptr == NULL ){ /* find which block to use */
	if( index > (p_clr->block_cnt-1)*CTABLE_SIZE && 
	    index < p_clr->block_cnt*CTABLE_SIZE ) /* is last block correct */
	    c_ptr = p_clr->cur_block;
	else
	    for( c_ptr = p_clr->clr_block, i = CTABLE_SIZE ; 
			i <= index ; i += CTABLE_SIZE)
		c_ptr = c_ptr->next;
    }

    i = index % CTABLE_SIZE;
    c_ptr->tbl[i].index = index;
    memcpy( (char *) &c_ptr->tbl[i].col, (char *) col, sizeof(PCOLOR));

    return 0;
}

/*
 *  ROUTINE:    DB_get_color_values
 *  PURPOSE:	This routine will retrieve 'col' from the color table block
 *		at the location 'index'.
 *		The location at which index is located is calculated by:
 *			block  = index / blocksize + 1
 *			offset = index % blocksize 
 *
 *		    where block is base 1, offset is base 0
 *
 *  RETURNS:
 *	0 if successful, 1 if error.
 */
int
DB_get_color_values( index, col)
int	index;
PCOLOR	*col;
{
    CTABLE_LIST *c_ptr;
    int i,nblock;
	/* has any color blocks been allocated yet ? */
    if(p_clr->clr_block == NULL ){
	return 1;
    }

    nblock = index / CTABLE_SIZE;

    c_ptr = p_clr->clr_block;

    for(i=0;i<nblock-1 && c_ptr != NULL;i++)
	c_ptr = c_ptr->next;

    if( c_ptr == NULL) return 1;

    i = index % CTABLE_SIZE;
    memcpy( (char *) col, (char *) &c_ptr->tbl[i].col, sizeof(PCOLOR));

    return 0;
}
