/*
 *  $Header: /auto/nfs/sequoia/u/home/rws/pextk/src/pextk_dls.c,v 1.1 1993/10/12 00:07:09 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                             pextk_dls.c                            *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */

/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk Display List routines.
 *	
 *
 *  Module Notes:
 *	Variables and their uses:
 *	    dls_root_node	Points to the linked list of display list
 *				'root' nodes. These 'root' nodes are those
 *				with an opcode of 'PEXtk_OP_DPY_LIST', and
 *				contains the 'name' of the display list in
 *				arg1.l. The pointer to the next display list
 *				root node is in 'arg3.ptr'.
 *				Note that every newly allocated display list
 *				header node becomes the dls_root_node.
 *	    dls_free_list	Memory list where all nodes are allocated
 *				as a 'block'.
 *	    dls_mem_node	Linked list of pointers to the allocated
 *				'dls_free_list'.
 *
 *	A Display List is stored based on a user given 'name', which cannot
 *	be zero.
 *	When a dpy_list is created with the Pcreate_dpy_list routine, the
 *	get_dls_node routine is called to either return the next free node,
 *	or to allocate a new block. The get_dls_node routine will update
 *	the dls_mem_node to point to any newly allocated memory block.
 *	The returned node will be added to the dls_root_node list.
 *
 *	A Display List is stored as a linked list of data blocks, 
 *	where a data block is a PEXtk_DLS_NODE (of the
 *	size PEXtk_DPY_LIST_SIZE). Each dpy_list header node points back
 *	to the previous header node with arg3.ptr.
 *
 *	The dls_root_node always points to the 'current' root node, not
 *	necessarily the first node allocated.
 *
 *	For a given display list 'name', the node has an opcode of
 *	PEXtk_OP_DPY_LIST, and arg1.l is 'name'. This indicates the start
 *	of that display list.
 *
 *	Routines in this module use the following naming format:
 *	    DLS		-	Stores the data for that command/op
 *	    DLI		-	Executes the command/op
 *
 *
 * 	dls_root_node		    Root node is the last allocated header node
 *
 *	      +-->  opcode	    PEXtk_OP_DPY_LIST (dpy list header node)
 *	      | +-  next	    following dpy list command/op
 *	      |	|   arg1.l	    name
 *	    +-|-|-  arg2.ptr        next dpy list header node
 *	 +--|-|-|-  arg3.ptr        previous dpy list header node
 *	 |  | |	+--->	opcode	    command/op
 *	 |  | |   +--	next
 *	 |  | |   |	arg1.()	    first arg
 *	 |  | |	  |	arg2.()	    second arg
 *	 |  | |	  |	arg3-7
 *	 |  | |	  +--->	opcode	    command/op
 *	 |  | |   +--	next
 *	 |  | |   |	arg1.()	    first arg
 *	 |  | |	  |	arg2.()	    second arg
 *	 |  | |	  |	arg3-7
 *	 |  | |	  .	.
 *	 |  | |	  .	.
 *	 |  | |	  +--->	opcode	    command/op
 *	 |  | |   +--	next
 *	 |  | |   |	arg1.()	    first arg
 *	 |  | |	  |	arg2.()	    second arg
 *	 |  | |	  |	arg3-7
 *	 |  | |   +--->	nil
 *	 |  | |   
 *	 |  | |	  next header
 *	 |  +-----> opcode	    PEXtk_OP_DPY_LIST (dpy list header node)
 *	 |    |     next	    following dpy list command/op
 *	 |    |     arg1.l	    name
 *	 |  +-|--   arg2.ptr        next dpy list header node
 *	 |  | +---  arg3.ptr        previous dpy list header node
 *	 |  |		opcode	    command/op
 *	 |  |     +--	next
 *	 |  |	  |	arg1.()	    int arg
 *	 |  |	  |	arg2.()	    second arg
 *	 |  |	  .     .
 *	 |  |	  .     .
 *	 |  |	  +---> opcode
 *	 |  |     +--	next
 *	 |  |	  |	arg1.()	    int arg
 *	 |  |	  |	arg2.()	    second arg
 *	 |  |	  +---> nil
 *	 |  |
 *	 +--+-----+-> nil
 *
 *  Module Contents:
 *
 *    int   PEXtk_create_dpy_list      Creates the display list
 *	
 */

#include <stdio.h>
#include <math.h>
#include <string.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#ifdef MPEX
#include <X11/MPEX.h>
#include <X11/MPEXproto.h>
#include <X11/MPEXlib.h>
#endif

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"
#include "pextk_ops.h"

    static int 		  dls_free_cnt;		/* avail number of nodes   */
    static DPY_LIST_NODE  *dls_root_node;	/* ptr to dls root node    */
    static DPY_LIST_NODE  *dls_free_list;	/* ptr to dls free list    */
    static DPY_LIST_NODE  *str_root_node;	/* ptr to struct root node */
    static DPY_LIST_NODE  *dpy_list_start;	/* first block in dpy list */
    static DPY_LIST_NODE  *cur_dls_node;
    static DPY_LIST_NODE  *get_dls_node( );
    static DPY_LIST_NODE  *DLS_find_dls_node( );
    static PTR_DLS_NODE	  *dls_mem_node;
    static int 		  dls_free_pnt3_cnt;	/* avail number of pnt3 nodes */
    static int 		  dls_free_pnt4_cnt;	/* avail number of pnt4 nodes */
    static DPY_POINT3_NODE *dls_free_pnt3_list; /* ptr to dls free list    */
    static DPY_POINT4_NODE *dls_free_pnt4_list; /* ptr to dls free list    */
    static DPY_POINT3_NODE *get_point3_node( );
    static DPY_POINT4_NODE *get_point4_node( );
    static PTR_POINT3_NODE *dls_mem_pnt3_node;
    static PTR_POINT4_NODE *dls_mem_pnt4_node;

#define SET_PCOLOR_DATA( col, ptr ) \
		    col.type = ptr->arg1.i; \
		    col.color.rgb.red   = ptr->arg2.f;\
		    col.color.rgb.green = ptr->arg3.f;\
		    col.color.rgb.blue  = ptr->arg4.f;

#define DLS_DEFAULT_VIEW_SIZE	100

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_init_dls      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes the display list routines.
 *
 *  USAGE:
 *	PEXtk_init_dls( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *
 *  NOTE:
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_init_dls(p_obj)
PEXTK_OBJECT		*p_obj;
{
    /*
     * Save the various object ptrs
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;

    dls_free_cnt	= 0;	
    dpy_list_start	= NULL;	/* not used */
    dls_root_node	= NULL;
    dls_free_list	= NULL;
    cur_dls_node  	= NULL;
    dls_mem_node	= NULL;
    dls_free_pnt3_cnt	= 0;	
    dls_free_pnt4_cnt	= 0;	
    dls_free_pnt3_list	= NULL;
    dls_free_pnt4_list	= NULL;
    dls_mem_pnt3_node	= NULL;
    dls_mem_pnt4_node	= NULL;

    /* Allocate the data for local data storage */
    if((p_dl->view_block = (PEXtk_VIEW *)malloc( 
	  sizeof(PEXtk_VIEW)*DLS_DEFAULT_VIEW_SIZE))
		== NULL ) return 1;
    p_dl->view_block_size = DLS_DEFAULT_VIEW_SIZE;

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_close_dls     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Close the display list routines.
 *
 *  USAGE:
 *	PEXtk_close_dls( )
 *
 *  DESCRIPTION:
 *
 *  NOTE:
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_close_dls()
{
    free(p_dl->view_block);
    p_dl->view_block = NULL;
    /* delete all display lists */
    /* TBD */
    return 0;
}

/*
 *  ROUTINE:    DLS_create_dpy_list
 *  PURPOSE:	Creates a display list
 *  RETURNS:	0 if successfull, 1 if error
 */
int
DLS_create_dpy_list( name )
int	name;
{
    DPY_LIST_NODE *ptr;

#define NO_DB_DLS
#ifndef NO_DB_DLS
		/* make sure this dpy list name doesn't exist already */
    	if( DLS_find_dls_node( name ) ) 
	    return 1;
#endif

    	if(( ptr = get_dls_node()) == NULL){
	    printf("ERROR... getting node for dpy list\n");
	    return 1; /* memory alloc error */
	}

	ptr->opcode   = PEXtk_OP_DPY_LIST;
	ptr->next     = NULL;
	ptr->arg1.l   = name;
	ptr->arg2.ptr = (PTR) dls_root_node;
	ptr->arg3.ptr = NULL;
	if( dls_root_node != NULL )
	    dls_root_node->arg3.ptr  = (PTR) ptr;
	dls_root_node = cur_dls_node = ptr;

    return 0;
}

/*
 *  ROUTINE:    DLS_delete_dpy_list
 *  PURPOSE:	Deletes a display list 
 */
int
DLS_delete_dpy_list( name )
int	name;
{
    DPY_LIST_NODE *ptr, *ptr2;

	if(( ptr = DLS_find_dls_node( name)) == NULL) /* get ptr to node */
	    return 1;	/* can't find dls node */
	/* delete this node */
	DLS_delete_dpy_list_node( ptr );

	while( ptr ){
	    switch( ptr->opcode ){	/* free pointers */
		default:
		    break;
	    }
	    ptr2 = ptr->next;
	    free_dli_node( ptr);
	    ptr = ptr2;
	}

    return 0;
}
/*
 *  ROUTINE:    DLS_delete_dpy_list_node
 *  PURPOSE:	Deletes a display list node
 */
int
DLS_delete_dpy_list_node( ptr )
DPY_LIST_NODE	*ptr;
{
	if( dls_root_node == ptr )
	    dls_root_node = (DPY_LIST_NODE *) ptr->arg2.ptr;
	if( ptr->arg3.ptr )
	    ((DPY_LIST_NODE *)ptr->arg3.ptr)->arg2.ptr = ptr->arg2.ptr;
	if( ptr->arg2.ptr )
	    ((DPY_LIST_NODE *)ptr->arg2.ptr)->arg3.ptr = ptr->arg3.ptr;
	ptr->next = dls_free_list;	/* put on free list */
	dls_free_list = ptr;
	dls_free_cnt++;
	return 0;
	
}

/*
 *  ROUTINE:    DLS_call_dpy_list
 *  PURPOSE:	Calls/Executes a display list 
 */
int
DLS_call_dpy_list( name )
int	name;
{
    DPY_LIST_NODE *ptr;

	if(( ptr = DLS_find_dls_node( name)) == NULL){ /* get ptr to node */
	    return 1;	/* can't find dls node */
	}

	DLI_call_dls( ptr);

    return 0;
}
/*
 *  ROUTINE:    get_dls_node
 *  PURPOSE:	Returns a pointer to the next available display list node.
 *  DESCRIPTION:  The number of available nodes is checked (dls_free_cnt)
 *		  and if there is no room left, then a new block of nodes
 *		  is allocated. If a new block is allocated, then the pointer
 *		  to the allocated node is placed in the DPY_ALLOC_NODES
 *		  structure, so that it may be freed when necessary.
 */
static
DPY_LIST_NODE *
get_dls_node( )
{
    int i;
    DPY_LIST_NODE *d_ptr;
    PTR_DLS_NODE  *p_ptr;

    if( ! dls_free_cnt ){
	if((dls_free_list = (DPY_LIST_NODE *)
	    malloc( sizeof(DPY_LIST_NODE)*DLS_NODE_LL_SIZE)) == NULL)
	    return NULL;
	/* get memory for the base/root node */
	if((p_ptr = (PTR_DLS_NODE *) malloc( sizeof(PTR_DLS_NODE))) == NULL)
	    return NULL;
	for( i = 0; i < DLS_NODE_LL_SIZE-1; i++) /* init next ptrs */
	    (dls_free_list+i)->next = (dls_free_list + i + 1);

				/* add node ptr to node ptr list */
	/* DB_add_dls_alloc_block( dls_free_list ); */
	p_ptr->ptr = dls_free_list;  /* set ptr to beginning of list */
	p_ptr->next = dls_mem_node;
	dls_mem_node = p_ptr;
	dls_free_cnt = DLS_NODE_LL_SIZE;
    }
    d_ptr = dls_free_list;		   /* return this node */
    dls_free_list = dls_free_list->next; /* point to next avail node */
    dls_free_cnt--;		   /* decrement the avail count */

    return d_ptr;
}

static
DPY_LIST_NODE *
DLS_find_dls_node( name )
int	name;
{

    DPY_LIST_NODE *dl;

    dl = dls_root_node;

    while( dl && dl->arg1.l != name )
	dl = (DPY_LIST_NODE *) dl->arg2.ptr;

    return dl;
}

/*
 *  ROUTINE:    get_point3_node
 *  PURPOSE:	Returns a pointer to the next available node for 3 vertices
 *  DESCRIPTION:  The number of available nodes is checked (dls_free_pnt3_cnt)
 *		  and if there is no room left, then a new block of nodes
 *		  is allocated. If a new block is allocated, then the pointer
 *		  to the allocated node is placed in the DPY_ALLOC_NODES
 *		  structure, so that it may be freed when necessary.
 */
static DPY_POINT3_NODE *
get_point3_node( )
{
    int i;
    DPY_POINT3_NODE *d_ptr;
    PTR_POINT3_NODE  *p_ptr;

    if( ! dls_free_pnt3_cnt ){
	if((dls_free_pnt3_list = (DPY_POINT3_NODE *)
	    malloc( sizeof(DPY_POINT3_NODE)*DLS_NODE_LL_SIZE)) == NULL)
	    return NULL;
	/* get memory for the base/root node */
	if((p_ptr = (PTR_POINT3_NODE *)malloc(sizeof(PTR_POINT3_NODE)))== NULL)
	    return NULL;
	for( i = 0; i < DLS_NODE_LL_SIZE-1; i++) /* init next ptrs */
	    (dls_free_pnt3_list+i)->next = (dls_free_pnt3_list + i + 1);

				/* add node ptr to node ptr list */
	/* DB_add_dls_alloc_block( dls_free_pnt3_list ); */
	p_ptr->ptr = dls_free_pnt3_list;  /* set ptr to beginning of list */
	p_ptr->next = dls_mem_pnt3_node;
	dls_mem_pnt3_node = p_ptr;
	dls_free_pnt3_cnt = DLS_NODE_LL_SIZE;
    }
    d_ptr = dls_free_pnt3_list;		   /* return this node */
    dls_free_pnt3_list = dls_free_pnt3_list->next;/* point to next avail node */
    dls_free_pnt3_cnt--;		   /* decrement the avail count */

    return d_ptr;
}

/*
 *  ROUTINE:    get_point4_node
 *  PURPOSE:	Returns a pointer to the next available node for 4 vertices
 *  DESCRIPTION:  The number of available nodes is checked (dls_free_pnt4_cnt)
 *		  and if there is no room left, then a new block of nodes
 *		  is allocated. If a new block is allocated, then the pointer
 *		  to the allocated node is placed in the DPY_ALLOC_NODES
 *		  structure, so that it may be freed when necessary.
 */
static DPY_POINT4_NODE *
get_point4_node( )
{
    int i;
    DPY_POINT4_NODE *d_ptr;
    PTR_POINT4_NODE  *p_ptr;

    if( ! dls_free_pnt4_cnt ){
	if((dls_free_pnt4_list = (DPY_POINT4_NODE *)
	    malloc( sizeof(DPY_POINT4_NODE)*DLS_NODE_LL_SIZE)) == NULL)
	    return NULL;
	/* get memory for the base/root node */
	if((p_ptr = (PTR_POINT4_NODE *)malloc(sizeof(PTR_POINT4_NODE)))== NULL)
	    return NULL;
	for( i = 0; i < DLS_NODE_LL_SIZE-1; i++) /* init next ptrs */
	    (dls_free_pnt4_list+i)->next = (dls_free_pnt4_list + i + 1);

				/* add node ptr to node ptr list */
	/* DB_add_dls_alloc_block( dls_free_pnt4_list ); */
	p_ptr->ptr = dls_free_pnt4_list;  /* set ptr to beginning of list */
	p_ptr->next = dls_mem_pnt4_node;
	dls_mem_pnt4_node = p_ptr;
	dls_free_pnt4_cnt = DLS_NODE_LL_SIZE;
    }
    d_ptr = dls_free_pnt4_list;		   /* return this node */
    dls_free_pnt4_list = dls_free_pnt4_list->next;/* point to next avail node */
    dls_free_pnt4_cnt--;		   /* decrement the avail count */

    return d_ptr;
}

/*
 *  ROUTINE:    DLS_add_view
 *  PURPOSE:	This routine will add view data into the view block.
 */
int
DLS_add_view( vw )
PEXtk_VIEW	*vw;
{
    int i;
	/* has the block been allocated yet ? */
    if(p_dl->view_block_cnt+1 >= p_dl->view_block_size ){
	if((p_dl->view_block = (PEXtk_VIEW *)realloc( p_dl->view_block,
	  sizeof(PEXtk_VIEW)*(p_dl->view_block_size+DLS_DEFAULT_VIEW_SIZE)))
		== NULL ) return 1;
        p_dl->view_block_size += PEXtk_DEFAULT_STRUCT_SIZE;
    }
    i = p_dl->view_block_cnt;
    p_dl->view_block[i].type = vw->type;
    p_dl->view_block[i].id = vw->id;
    p_dl->view_block[i].width = vw->width;
    p_dl->view_block[i].height = vw->height;
    p_dl->view_block[i].left = vw->left;
    p_dl->view_block[i].right = vw->right;
    p_dl->view_block[i].bottom = vw->bottom;
    p_dl->view_block[i].top = vw->top;
    p_dl->view_block[i].near = vw->near;
    p_dl->view_block[i].far = vw->far;
    p_dl->view_block[i].aspect = vw->aspect;
    p_dl->view_block[i].fov = vw->fov;

    p_dl->view_block_cnt++;

    return 0;

}

/*
 *  ROUTINE:    DLI_call_dls
 *  PURPOSE:	Calls/Executes a display list
 *  DESCRIPTION:  This routine is called by DLS_call_dpy_list, and it is
 *		  passed the beginning node (display list header node) of
 *		  the display list 'name'.
 */
DLI_call_dls( ptr )
DPY_LIST_NODE	*ptr;
{
    PCOLOR col;
    while (ptr){
      switch( ptr->opcode & PEXtk_FUNC_MASK){
	case PEXtk_FUNC_NO_OP:
	  switch( ptr->opcode ){
	     case PEXtk_OP_DPY_LIST:
			break;
	    }
	    break;
	case PEXtk_FUNC_XFORM:
	  switch( ptr->opcode ){
	     case PEXtk_OP_ROT:
		    Protate( ptr->arg1.f, ptr->arg2.c, ptr->arg3.i );
		break;
	     case PEXtk_OP_ROT_V0:
		    Protate_vector( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f,
			ptr->arg4.f, ptr->arg5.f, ptr->arg6.f, ptr->arg7.f, 0);
		break;
	     case PEXtk_OP_ROT_V1:
		    Protate_vector( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f,
			ptr->arg4.f, ptr->arg5.f, ptr->arg6.f, ptr->arg7.f, 1);
		break;
	     case PEXtk_OP_ROT_X:
#define DEBUG
#ifdef DEBUG
printf(" rot_x %f, %d\n",ptr->arg2.f, ptr->arg1.i);
#endif
		    Protate_x( ptr->arg2.f, ptr->arg1.i );
		break;
	     case PEXtk_OP_ROT_Y:
		    Protate_y( ptr->arg2.f, ptr->arg1.i );
		break;
	     case PEXtk_OP_ROT_Z:
		    Protate_z( ptr->arg2.f, ptr->arg1.i );
		break;
	     case PEXtk_OP_SCALE:
		    Pscale( ptr->arg2.f, ptr->arg3.f, ptr->arg4.f, ptr->arg1.i);
		break;
	     case PEXtk_OP_TRAN:
#ifdef DEBUG
printf(" tran %f, %f, %f, %d\n",ptr->arg2.f,ptr->arg3.f,ptr->arg4.f, 
		ptr->arg1.i);
#endif
		    Ptranslate( ptr->arg2.f, ptr->arg3.f, ptr->arg4.f, 
			ptr->arg1.i);
		break;
	     case PEXtk_OP_TRAN_X:
#ifdef DEBUG
printf(" tran_x %f, %d\n",ptr->arg2.f, ptr->arg1.i);
#endif
		    Ptranslate_x( ptr->arg2.f, ptr->arg1.i );
		break;
	     case PEXtk_OP_TRAN_Y:
#ifdef DEBUG
printf(" tran_y %f, %d\n",ptr->arg2.f, ptr->arg1.i);
#endif
		    Ptranslate_y( ptr->arg2.f, ptr->arg1.i );
		break;
	     case PEXtk_OP_TRAN_Z:
#ifdef DEBUG
printf(" tran_z %f, %d\n",ptr->arg2.f, ptr->arg1.i);
#endif
		    Ptranslate_z( ptr->arg2.f, ptr->arg1.i );
		break;
	    }
	    break;
	case PEXtk_FUNC_PRIM:
	  switch( ptr->opcode ){
	     case PEXtk_OP_POLY:
#ifdef DEBUG
printf(" Ppoly %x, %d\n",ptr->arg2.ptr, ptr->arg1.i);
#endif
		    Ppoly( ptr->arg1.i, (Fcoord3D *) ptr->arg2.ptr );
		break;
	    }
	    break;
	case PEXtk_FUNC_ATTRIB:
	  switch( ptr->opcode ){
	     case PEXtk_OP_BACKFACE:
		    Pbackface( ptr->arg1.i);
		break;
	     case PEXtk_OP_COLOR:
		    SET_PCOLOR_DATA( col, ptr);
		    Pcolor( &col);
#ifdef DEBUG
printf(" Pcolor %f, %f, %f\n",col.color.rgb.red,col.color.rgb.green,col.color.rgb.blue);
#endif
		break;
	     case PEXtk_OP_COLOR_HSV:
		    Pcolor_hsv( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f);
		break;
	     case PEXtk_OP_COLOR_RGB:
		    Pcolor_rgb( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f);
		break;
	     case PEXtk_OP_LINE_CLR:
		    SET_PCOLOR_DATA( col, ptr);
#ifdef DEBUG
printf(" Pline_color %f, %f, %f\n",col.color.rgb.red,col.color.rgb.green,col.color.rgb.blue);
#endif
		    Pline_color( &col);
		break;
	     case PEXtk_OP_LINE_CLR_HSV:
		    Pline_color_hsv( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f);
		break;
	     case PEXtk_OP_LINE_CLR_RGB:
		    Pline_color_rgb( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f);
		break;
	     case PEXtk_OP_LINE_TYPE:
		    Pline_type( ptr->arg1.i);
		break;
	     case PEXtk_OP_LINE_WIDTH:
		    Pline_width( ptr->arg1.f);
		break;
	     case PEXtk_OP_MAKE_LABEL:
		    Pmake_label( ptr->arg1.i);
		break;
	     case PEXtk_OP_MRKR_CLR:
		    SET_PCOLOR_DATA( col, ptr);
		    Pmarker_color( &col);
		break;
	     case PEXtk_OP_MRKR_CLR_HSV:
		    Pmarker_color_hsv( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f);
		break;
	     case PEXtk_OP_MRKR_CLR_RGB:
		    Pmarker_color_rgb( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f);
		break;
	     case PEXtk_OP_TEXT_CLR:
		    SET_PCOLOR_DATA( col, ptr);
		    Ptext_color( &col);
		break;
	     case PEXtk_OP_TEXT_CLR_HSV:
		    Ptext_color_hsv( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f);
		break;
	     case PEXtk_OP_TEXT_CLR_RGB:
		    Ptext_color_rgb( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f);
		break;
	    }
	    break;
	case PEXtk_FUNC_STATE:
	  switch( ptr->opcode ){
	     case PEXtk_OP_BUFFER_MODE:
		    Pbuffer_mode( ptr->arg1.i);
		break;
	     case PEXtk_OP_MATRIX_MODE:
		    Pmatrix_mode( ptr->arg1.i);
		break;
	     case PEXtk_OP_PUSH_ATTR:
		    Ppush_attributes( );
		break;
	     case PEXtk_OP_POP_ATTR:
		    Ppush_attributes( );
		break;
	     case PEXtk_OP_PUSH_MTX:
		    Ppush_matrix( );
		break;
	     case PEXtk_OP_POP_MTX:
		    Ppop_matrix( );
		break;
	     case PEXtk_OP_PUSH_NAME:
		    Ppush_name( ptr->arg1.i );
		break;
	     case PEXtk_OP_POP_NAME:
		    Ppop_name( );
		break;
	     case PEXtk_OP_PUSH_VIEW:
		    Ppush_view( );
		break;
	     case PEXtk_OP_POP_VIEW:
		    Ppush_view( );
		break;
	     case PEXtk_OP_PUSH_VPT:
		    Ppush_viewport( );
		break;
	     case PEXtk_OP_POP_VPT:
		    Ppush_viewport( );
		break;
	     case PEXtk_OP_SET_DRAWABLE:
		    Pset_drawable( ptr->arg1.l);
		break;
	     case PEXtk_OP_SHADE_MODE:
		    Pshade_mode( ptr->arg1.i);
		break;
	     case PEXtk_OP_Z_BUFFER:
		    Pzbuffer( ptr->arg1.i);
		break;
	    }
	    break;
	case PEXtk_FUNC_EXEC:
	  switch( ptr->opcode ){
	     case PEXtk_OP_CALL_STRUCT:
#ifdef DEBUG
printf(" call_struct %d\n",ptr->arg1.i);
#endif
		    Pcall_struct( ptr->arg1.i);
		break;
	     case PEXtk_OP_EXEC_STRUCT:
#ifdef DEBUG
printf(" exec_struct %d\n",ptr->arg1.i);
#endif
		    Pexec_struct( ptr->arg1.i);
		break;
	    }
	    break;
	case PEXtk_FUNC_VIEW:
	  switch( ptr->opcode ){
	     case PEXtk_OP_VIEW_LOOKAT:
		  { PEXtk_VIEW *vw;
		    /*  note the argument names don't correspond with the
			lookat view argument names */
		    vw = &(p_dl->view_block[ptr->arg1.i]);
		    Plookat_view( vw->left, vw->right, vw->bottom,
			vw->top, vw->near, vw->far, vw->fov);
		  }
		break;
	     case PEXtk_OP_VPT:
#ifdef DEBUG
printf(" vpt %d, %d, %d, %d\n",ptr->arg1.i, ptr->arg2.i, ptr->arg3.i, 
			ptr->arg4.i);
#endif
		    Pviewport( ptr->arg1.i, ptr->arg2.i, ptr->arg3.i, 
			ptr->arg4.i);
		break;
	     case PEXtk_OP_VIEW_ORTHO:
		    Portho_view( ptr->arg1.i, ptr->arg2.f, ptr->arg3.f, 
			ptr->arg4.f, ptr->arg5.f, ptr->arg6.f, ptr->arg7.f);
		break;
	     case PEXtk_OP_VIEW_ORTHO_2D:
		    Portho_2d_view( ptr->arg1.i, ptr->arg2.f, ptr->arg3.f, 
			ptr->arg4.f, ptr->arg5.f);
		break;
	     case PEXtk_OP_VIEW_PERSP:
#ifdef DEBUG
printf(" persp %d, %f, %f, %f, %f\n",ptr->arg1.i, ptr->arg2.f, ptr->arg3.f, 
			ptr->arg4.f,ptr->arg5.f);
#endif
		    Ppersp_view( ptr->arg1.i, ptr->arg2.f, ptr->arg3.f, 
			ptr->arg4.f, ptr->arg5.f );
		break;
	     case PEXtk_OP_VIEW_POLAR:
		    Ppolar_view( ptr->arg1.f, ptr->arg2.f, ptr->arg3.f, 
			ptr->arg4.f);
		break;
	     case PEXtk_OP_VIEW_WINDOW:
		    Pwindow_view( ptr->arg1.i, ptr->arg2.f, ptr->arg3.f, 
			ptr->arg4.f, ptr->arg5.f, ptr->arg6.f, ptr->arg7.f);
		break;
	    }
	    break;
	default:
	    break;
      } /* end switch ( opcode & MASK ) */
      ptr = ptr->next;
    } /* end while */
}

DLI_display( name )
int	  name;
{
}

DLS_call_dls( name )
int	name;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = PEXtk_OP_CALL_DLS;
	ptr->next     = NULL;
	ptr->arg1.l   = name;
	ptr->arg2.ptr = (PTR) dls_root_node;
	ptr->arg3.ptr = NULL;
	if( dls_root_node != NULL )
	    dls_root_node->arg3.ptr  = (PTR) ptr;
	dls_root_node = cur_dls_node = ptr;

    return 0;
}

DLS_xform( op, flag, arg1, arg2, arg3 )
int	op;
int	flag;
float   arg1, arg2, arg3;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.l   = flag;
	switch( op ){
	    case PEXtk_OP_ROT:	/* not used */
		break;
	    case PEXtk_OP_ROT_X:
	    case PEXtk_OP_ROT_Y:
	    case PEXtk_OP_ROT_Z:
		ptr->arg2.f   = arg1;
		break;
	    case PEXtk_OP_TRAN_X:
	    case PEXtk_OP_TRAN_Y:
	    case PEXtk_OP_TRAN_Z:
		ptr->arg2.f   = arg1;
		break;
	    case PEXtk_OP_TRAN:
	    case PEXtk_OP_SCALE:
		ptr->arg2.f   = arg1;
		ptr->arg3.f   = arg2;
		ptr->arg4.f   = arg3;
		break;
	}

    return 0;
}

DLS_rot_vector( ox, oy, oz, vx, vy, vz, a, flag )
float   ox, oy, oz, vx, vy, vz;
Angle   a;
int	flag;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	if( flag == 1 ){
	    ptr->opcode   = PEXtk_OP_ROT_V1;
	} else {
	    ptr->opcode   = PEXtk_OP_ROT_V0;
	}
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.f   = ox;
	ptr->arg2.f   = oy;
	ptr->arg3.f   = oz;
	ptr->arg4.f   = vx;
	ptr->arg5.f   = vy;
	ptr->arg6.f   = vz;
	ptr->arg7.f   = a;

    return 0;
}

DLS_view_box( op, arg1, arg2, arg3, arg4 ) /* clip_rect, viewport */
int	op;
int     arg1, arg2, arg3, arg4;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_CLIP_RECT:
	    case PEXtk_OP_VPT:
		ptr->arg1.l   = arg1;
		ptr->arg2.l   = arg2;
		ptr->arg3.l   = arg3;
		ptr->arg4.l   = arg4;
		break;
	}

    return 0;
}

DLS_view_proj( op, id, arg1, arg2, arg3, arg4, arg5, arg6 ) /* P_..view */
int	op;
int     id;
float   arg1, arg2, arg3, arg4, arg5, arg6;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.l   = id;
	ptr->arg2.f   = arg1;
	ptr->arg3.f   = arg2;
	ptr->arg4.f   = arg3;
	ptr->arg5.f   = arg4;
	switch( op ){		/* for these, two more arg's are given */
	    case PEXtk_OP_VIEW_ORTHO:
	    case PEXtk_OP_VIEW_WINDOW:
		ptr->arg6.f   = arg5;
		ptr->arg7.f   = arg6;
		break;
	}

    return 0;
}

/*
 *  ROUTINE:	DLS_view_lookat
 *  PURPOSE:	Stores data for lookat view routine.
 *  NOTES:	This routine will be changed to reference a DLS table
 *		instead of storing the data in a node list.
 */
DLS_view_lookat( vx, vy, vz, px, py, pz, twist )
float   vx, vy, vz, px, py, pz, twist;
{
    DPY_LIST_NODE *ptr;
    PEXtk_VIEW vw;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = PEXtk_OP_VIEW_LOOKAT;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	DLS_add_view( &vw );
	vw.id      = 0;
	vw.type    = PEXtk_VIEW_LOOKAT;
	vw.left	   = vx;
	vw.right   = vy;
	vw.bottom  = vz;
	vw.top     = px;
	vw.near    = py;
	vw.far     = pz;
	vw.fov     = twist;
	if(DLS_add_view( &vw ))
	    return 1;	/* memory alloc error */
	ptr->arg1.i = p_dl->view_block_cnt-1;

    return 0;
}

/*
 *  ROUTINE:	DLS_view_polar
 *  PURPOSE:	Stores data for polar view routine.
 */
DLS_view_polar( dist, azim, inc, twist)
float dist, azim, inc, twist;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = PEXtk_OP_VIEW_POLAR;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.f   = dist;
	ptr->arg2.f   = azim;
	ptr->arg3.f   = inc;
	ptr->arg4.f   = twist;

    return 0;
}

DLS_view_mode( op, i_arg, f_arg1, f_arg2, f_arg3)
int	op;
int     i_arg;
float   f_arg1, f_arg2, f_arg3;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_DEPTH:
		ptr->arg1.f   = f_arg1;
		ptr->arg2.f   = f_arg2;
		break;
	    case PEXtk_OP_DEPTH_CUE:
		ptr->arg1.l   = i_arg;
		ptr->arg2.f   = f_arg1;
		ptr->arg3.f   = f_arg2;
		break;
	}

    return 0;
}

DLS_state( op, i_arg)
int	op;
int     i_arg;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_BUFFER_MODE:
		ptr->arg1.i   = i_arg;
		break;
	    case PEXtk_OP_MATRIX_MODE:
		ptr->arg1.i   = i_arg;
		break;
	    case PEXtk_OP_SET_DRAWABLE:
		ptr->arg1.l   = i_arg;
		break;
	    case PEXtk_OP_SHADE_MODE:
		ptr->arg1.l   = i_arg;
		break;
	    case PEXtk_OP_Z_BUFFER:
		ptr->arg1.i   = i_arg;
		break;
	}

    return 0;
}

DLS_set_drawable( op, i_arg)
int	op;
long     i_arg;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.l   = i_arg;

    return 0;
}

DLS_exec( op, i_arg)
int	op;
int     i_arg;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_CALL_STRUCT:
		ptr->arg1.l   = i_arg;
		break;
	    case PEXtk_OP_EXEC_STRUCT:
		ptr->arg1.l   = i_arg;
		break;
	}

    return 0;
}

DLS_attrib( op, i_arg)
int	op;
int	i_arg;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_BACKFACE:
	    case PEXtk_OP_MAKE_LABEL:
	    case PEXtk_OP_MRKR_SCALE:
	    case PEXtk_OP_MRKR_TYPE:
	    case PEXtk_OP_LD_IDENT:
	    case PEXtk_OP_LINE_TYPE:
	    case PEXtk_OP_LINE_WIDTH:
	    case PEXtk_OP_PICK_ID:
	    case PEXtk_OP_SET_VIEW:
	    case PEXtk_OP_TEXT_FNT:
	    case PEXtk_OP_TEXT_PATH:
		ptr->arg1.i   = i_arg;
		break;
	}

    return 0;
}

DLS_attrib_f( op, f_arg)
int	op;
float	f_arg;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_MRKR_SCALE:
		ptr->arg1.f   = f_arg;
		break;
	}

    return 0;
}

/*
 *  ROUTINE:	DLS_matrix
 *  PURPOSE:	For use by DLS routines that use matrices as arguments
 */
DLS_matrix( op, mat, arg1, arg2)
int    op;
Matrix mat;
int    arg1, arg2;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.i   = arg1;
	ptr->arg2.i   = arg2;

	/* store the matrix */
	/* TBD */

    return 0;
}

DLS_dither( op, i_arg1, i_arg2)
int	op;
int	i_arg1;
int	i_arg2;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = PEXtk_OP_DITHER;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.i   = i_arg1;
	ptr->arg2.i   = i_arg2;

    return 0;
}

DLS_transparency( op, i_arg, f_arg)
int	op;
int	i_arg;
float	f_arg;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = PEXtk_OP_TRANSPARENCY;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.i   = i_arg;
	ptr->arg2.f   = f_arg;

    return 0;
}

DLS_poly( op, n, pts)			/* for poly, poly_line */
int      op;
int	 n;
Fcoord3D *pts;
{
    DPY_LIST_NODE *ptr;
    DPY_POINT3_NODE *pt3;
    DPY_POINT4_NODE *pt4;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.i   = n;
	/* Get pointer to 'n' 3D points */
	if( n == 3 ){
	    pt3 = get_point3_node();
	    ptr->arg2.ptr = (PTR) pt3->pts;
	} else if( n == 4 ){
	    pt4 = get_point4_node();
	    ptr->arg2.ptr = (PTR) pt4->pts;
	} else
	    ptr->arg2.ptr = (PTR) malloc( n*sizeof(PEXCoord));
	if( ptr->arg2.ptr != NULL )
	    memcpy((char *)ptr->arg2.ptr,(char *)pts,n*sizeof(PEXCoord));
	else
	    return 1;

    return 0;
}

DLS_attrib_text( op, f_arg1, f_arg2)	/* for text_sz, text_orien */
int	op;
float	f_arg1, f_arg2;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.f   = f_arg1;
	ptr->arg2.f   = f_arg2;

    return 0;
}

DLS_text( op, x, y, z, string)	/* for text, text_annotation */
int     op;
float	x, y, z;
char    *string;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.f   = x;
	ptr->arg2.f   = y;
	ptr->arg3.f   = z;
	if(( ptr->arg4.ptr = (PTR) malloc(strlen(string))) == NULL)
	    return 1;
	strcpy(ptr->arg4.ptr, string);	/* CHANGE to pointer in DLS buffer */

    return 0;
}

DLS_attrib_clr( op, col)
int	op;
PCOLOR	*col;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_COLOR:
	    case PEXtk_OP_COLOR_SPEC:
	    case PEXtk_OP_LIGHT_CLR:
	    case PEXtk_OP_LINE_CLR:
	    case PEXtk_OP_MRKR_CLR:
	    case PEXtk_OP_TEXT_CLR:
		ptr->arg1.l   = col->type;
		ptr->arg2.f   = col->color.rgb.red;
		ptr->arg3.f   = col->color.rgb.green;
		ptr->arg4.f   = col->color.rgb.blue;
		break;
	}

    return 0;
}

DLS_attrib_clr_3f( op, arg1, arg2, arg3 )
int	op;
float   arg1, arg2, arg3;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_COLOR_RGB:
	    case PEXtk_OP_COLOR_HSV:
	    case PEXtk_OP_COLOR_SPEC_RGB:
	    case PEXtk_OP_COLOR_SPEC_HSV:
	    case PEXtk_OP_LIGHT_CLR_RGB:
	    case PEXtk_OP_LIGHT_CLR_HSV:
	    case PEXtk_OP_LINE_CLR_RGB:
	    case PEXtk_OP_LINE_CLR_HSV:
	    case PEXtk_OP_MRKR_CLR_RGB:
	    case PEXtk_OP_MRKR_CLR_HSV:
	    case PEXtk_OP_TEXT_CLR_RGB:
	    case PEXtk_OP_TEXT_CLR_HSV:
		ptr->arg1.f   = arg1;
		ptr->arg2.f   = arg2;
		ptr->arg3.f   = arg3;
		break;
	}

    return 0;
}

DLS_attrib_clr_ind( op, i_arg )
int	op;
int	i_arg;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_COLOR_IND:
	    case PEXtk_OP_COLOR_SPEC_IND:
	    case PEXtk_OP_LIGHT_CLR_IND:
	    case PEXtk_OP_LINE_CLR_IND:
	    case PEXtk_OP_MRKR_CLR_IND:
	    case PEXtk_OP_TEXT_CLR_IND:
		ptr->arg1.l   = i_arg;
		break;
	}

    return 0;
}

DLS_bld( op, arg1, arg2, arg3, arg4 )
int	op;
float   arg1, arg2, arg3, arg4;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_VERTEX_CLR:
	    case PEXtk_OP_VERTEX_NORM:
	    case PEXtk_OP_POLY_NORMAL:
		ptr->arg1.f   = arg1;
		ptr->arg2.f   = arg2;
		ptr->arg3.f   = arg3;
		break;
	}

    return 0;
}

DLS_stack( op )
int	op;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_POP_ATTR:
	    case PEXtk_OP_POP_VIEW:
	    case PEXtk_OP_POP_MTX:
	    case PEXtk_OP_POP_VPT:
	    case PEXtk_OP_PUSH_ATTR:
	    case PEXtk_OP_PUSH_VIEW:
	    case PEXtk_OP_PUSH_MTX:
	    case PEXtk_OP_PUSH_VPT:
		break;
	}

    return 0;
}

DLS_name_stack( op, value )
int	op;
int	value;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	switch( op ){
	    case PEXtk_OP_PUSH_NAME:
		ptr->arg1.i   = value;
		break;
	}

    return 0;
}

DLS_light( n, lgt)
int		n;
PEXtk_LIGHT	*lgt;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = PEXtk_OP_LIGHT;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.l   = n;
	ptr->arg2.ptr = (PTR) lgt;  /* CHANGE to pointer to DLS light table */

    return 0;
}

DLS_light_ambient( amb)
float	amb;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = PEXtk_OP_LIGHT_AMB;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.f   = amb;

    return 0;
}

DLS_refl_prop( arg1, arg2)
float	arg1, arg2;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = PEXtk_OP_REFL_PROP;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.f   = arg1;
	ptr->arg2.f   = arg2;

    return 0;
}

DLS_ext( op, arg1)		/* support for DEC extensions */
int		op;
unsigned long	arg1;
{
    DPY_LIST_NODE *ptr;

    	if(( ptr = get_dls_node()) == NULL)
	    return 1; /* memory alloc error */

	ptr->opcode   = op;
	ptr->next     = NULL;
	cur_dls_node->next = ptr;
	cur_dls_node  = ptr;
	ptr->arg1.l   = arg1;

    return 0;
}

free_dli_node( ptr )
DPY_LIST_NODE	*ptr;
{
    ptr->next		= dls_free_list;
    dls_free_list	= ptr;
    dls_free_cnt++;

    return ;
}

#define DEBUG
	char *DLS_get_opcode_string();

DLS_print_dpy_list()
{
#ifdef DEBUG
    int i,j;
    DPY_LIST_NODE *ptr, *ptr1;
    DPY_POINT3_NODE *pt3;
    DPY_POINT4_NODE *pt4;
    float *ptf;

    i = 0;
    ptr = dls_root_node;
	printf("Header nodes\n");
    while( ptr ){
	if( i % 20 == 0 )
	  printf(" addr       opcode     next      arg1       arg2     arg3\n");
      printf(" %8x  %8x %8x %8d     %8x %8x\n",
	ptr, ptr->opcode, ptr->next, ptr->arg1.l, ptr->arg2.ptr,ptr->arg3.ptr);
      ptr = (DPY_LIST_NODE *)ptr->arg2.ptr;
      i++;
    }
    ptr = ptr1 = dls_root_node;
    i = 0;
    while( ptr1 ){
      printf(" Display List %d\n",ptr1->arg1.l);
      while( ptr ){
	if( i % 20 == 0 )
	  printf(" addr       opcode     next      arg1       arg2     arg3\n");
#define USE_STR
#ifdef USE_STR
      printf(" %8x  %s  %8x %8d     %8x %8x\n",
	ptr, DLS_get_opcode_string(ptr->opcode), ptr->next, ptr->arg1.l, 
		ptr->arg2.ptr, ptr->arg3.ptr);
#else
      printf(" %8x  %8x %8x %8d     %8x %8x\n",
	ptr, ptr->opcode, ptr->next, ptr->arg1.l, ptr->arg2.ptr,ptr->arg3.ptr);
#endif
	if(ptr->opcode == PEXtk_OP_POLY){
	    switch( ptr->arg1.i ){
		case 3:

	    	    for(pt3= (DPY_POINT3_NODE *)ptr->arg2.ptr,j=0;j<3;j++)
			printf(" %f %f %f\n",
				pt3->pts[j].x, pt3->pts[j].y,
				pt3->pts[j].z);
		    break;
		case 4:
	    	    for(pt4= (DPY_POINT4_NODE *)ptr->arg2.ptr,j=0;j<4;j++)
			printf(" %f %f %f\n",
				pt4->pts[j].x, pt4->pts[j].y,
				pt4->pts[j].z);
		    break;
		default:
	    	    for(ptf= (float *)ptr->arg2.ptr,j=0;j<ptr->arg1.i;j++){
			printf(" %f", ptf); ptf++;
			printf(" %f", ptf); ptf++;
			printf(" %f\n", ptf); ptf++;
		    }
		    break;
	    }
	}
       ptr = ptr->next;
       i++;
       }
      ptr1 = (DPY_LIST_NODE *)ptr1->arg2.ptr;
      ptr = ptr1;
    }
#endif
}

char *
DLS_get_opcode_string(opcode)
int opcode;
{
    static dummy_string[11];

      switch( opcode & PEXtk_FUNC_MASK){
	case PEXtk_FUNC_NO_OP:
	  switch( opcode ){
	     case PEXtk_OP_DPY_LIST:
			break;
	    }
	    break;
	case PEXtk_FUNC_XFORM:
	  switch( opcode ){
	     case PEXtk_OP_ROT:
		    return "  rotate  ";
	     case PEXtk_OP_ROT_V0:
		    return " rotate_v0";
	     case PEXtk_OP_ROT_V1:
		    return " rotate_v1";
	     case PEXtk_OP_ROT_X:
		    return " rotate_x ";
	     case PEXtk_OP_ROT_Y:
		    return " rotate_y ";
	     case PEXtk_OP_ROT_Z:
		    return " rotate_z ";
	     case PEXtk_OP_SCALE:
		    return "  scale   ";
	     case PEXtk_OP_TRAN:
		    return " trans    ";
	     case PEXtk_OP_TRAN_X:
		    return "  tran_x  ";
	     case PEXtk_OP_TRAN_Y:
		    return "  tran_y  ";
	     case PEXtk_OP_TRAN_Z:
		    return "  tran_z  ";
	    }
	    break;
	case PEXtk_FUNC_PRIM:
	  switch( opcode ){
	     case PEXtk_OP_POLY:
		    return " Ppoly    ";
	    }
	    break;
	case PEXtk_FUNC_ATTRIB:
	  switch( opcode ){
	     case PEXtk_OP_BACKFACE:
		    return " backface ";
	     case PEXtk_OP_COLOR:
		    return " Pcolor   ";
	     case PEXtk_OP_COLOR_HSV:
		    return " Pcolor_h ";
	     case PEXtk_OP_COLOR_RGB:
		    return " Pcolor_r ";
	     case PEXtk_OP_LINE_CLR:
		    return " Pline_clr";
	     case PEXtk_OP_LINE_CLR_HSV:
		    return " Pln_clr_h";
	     case PEXtk_OP_LINE_CLR_RGB:
		    return " Pln_clr_r";
	     case PEXtk_OP_LINE_WIDTH:
		    return " line_widt";
	     case PEXtk_OP_MRKR_CLR:
		    return " mrkr_clr ";
	     case PEXtk_OP_MRKR_CLR_HSV:
		    return " mrkr_hsv ";
	     case PEXtk_OP_MRKR_CLR_RGB:
		    return " mrkr_rgb ";
	     case PEXtk_OP_TEXT_CLR:
		    return " text_clr ";
	     case PEXtk_OP_TEXT_CLR_HSV:
		    return " text_hsv ";
	     case PEXtk_OP_TEXT_CLR_RGB:
		    return " text_rgb ";
	    }
	    break;
	case PEXtk_FUNC_STATE:
	  switch( opcode ){
	     case PEXtk_OP_BUFFER_MODE:
		    return " buf_mode ";
	     case PEXtk_OP_MATRIX_MODE:
		    return " mat_mode ";
	     case PEXtk_OP_SET_DRAWABLE:
		    return " set_draw ";
	     case PEXtk_OP_SHADE_MODE:
		    return " shd_mode ";
	     case PEXtk_OP_Z_BUFFER:
		    return " z_buffer ";
	     case PEXtk_OP_PUSH_MTX:
		    return " push_mtx ";
	     case PEXtk_OP_POP_MTX:
		    return " pop_mtx  ";
	    }
	    break;
	case PEXtk_FUNC_EXEC:
	  switch( opcode ){
	     case PEXtk_OP_CALL_STRUCT:
		    return " call_str ";
	     case PEXtk_OP_EXEC_STRUCT:
		    return " exec_str ";
	    }
	    break;
	case PEXtk_FUNC_VIEW:
	  switch( opcode ){
	     case PEXtk_OP_VIEW_LOOKAT:
		    return " lookat   ";
	     case PEXtk_OP_VIEW_ORTHO:
		    return " ortho    ";
	     case PEXtk_OP_VIEW_ORTHO_2D:
		    return " ortho_2d ";
	     case PEXtk_OP_VIEW_PERSP:
		    return " persp    ";
	     case PEXtk_OP_VIEW_POLAR:
		    return " polar    ";
	     case PEXtk_OP_VPT:
		    return " viewport ";
	     case PEXtk_OP_VIEW_WINDOW:
		    return " window   ";
	    }
	    break;
	default:
	    break;
      } /* end switch ( opcode & MASK ) */

    sprintf( dummy_string, " %8x",opcode);
    return (char *)dummy_string;
}
