/*
 *  $Header: /auto/nfs/sequoia/u/home/rws/pextk/src/pextk_mtx.c,v 1.1 1993/10/12 00:07:09 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                          pextk_mtx.c                               *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk Matrix routines
 *
 *  Module Notes:
 *
 *  Module Contents:
 *
 *    void  Pget_identity_matrix  Copies a the 4x4 identity matrix into mat.
 *    void  Pget_inverse_matrix   Gets inverse of the current transformation
 *    void  Pget_matrix       Sends the stack's top 4x4 transformation matrix
 *    void  Pload_identity    Loads a 4x4 identity matrix onto the matrix sta
 *    void  Pload_matrix      Copies a 4x4 transformation matrix onto the mat
 *    void  Pmultiply_matrix  Mults a 4x4 matrix with the top of the stack
 *    void  Ppop_matrix       Pops the matrix stack. Only Immediate mode.
 *    void  Ppush_matrix      Pushes the matrix stack. Only Immediate mode.
 *    void  Protate           Calcs the x,y, or z rot. matrix and concats
 *    void  Protate_vector    Calcs the rotation matrix for rotation about line
 *    void  Protate_x         Calcs the x rotation matrix and concats on stk
 *    void  Protate_y         Calcs the y rotation matrix and concats on stk
 *    void  Protate_z         Calcs the z rotation matrix and concats on stk
 *    void  Pscale            Calcs the x,y,z scaling matrix and concats
 *    void  Ptranslate        Calcs the x,y,z trans. matrix and concats
 *    void  Ptranslate_x      Calcs the x trans. matrix and concats on stk
 *    void  Ptranslate_y      Calcs the y trans. matrix and concats on stk
 *    void  Ptranslate_z      Calcs the z trans. matrix and concats on stk
 *    int   PEXtk_Update_Matrix   Sets the global transform matrix
 *
 *
 *  Module Notes:
 *    1. The global transformation matrix is cached in immediate mode.
 *	 It is sent to the server when a geometry object, view, or
 *	 light is created/modified. The p_state->dirty_mask will have
 *	 the PEXtk_DIRTY_MTX bit set when the matrix has not been sent.
 *
 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"
#include "pextk_ops.h"

        /* static Matrix           mat_stk[32]; */
        static Matrix           *mat_stk;
        static PEXMatrix 	local_mtx;
        static Matrix           t, s, rx, ry, rz;
	static Matrix ident_mat  =     {{1.0, 0.0, 0.0, 0.0},
                                        {0.0, 1.0, 0.0, 0.0},
                                        {0.0, 0.0, 1.0, 0.0},
                                        {0.0, 0.0, 0.0, 1.0}};

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_init_mtx      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes matrix pointers.
 *
 *  USAGE:
 *	PEXtk_init_mtx( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *	This routine initializes pointers for p_build and p_state.
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
PEXtk_init_mtx(p_obj)
PEXTK_OBJECT		*p_obj;
{
    /*
     * Save the various object pointers
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;
    p_clr	= p_obj->clr_tbl;
    p_x_data	= p_obj->x_data;

    dpy = p_x_data->dpy;
    if(( mat_stk = (Matrix *) malloc(sizeof(Matrix)*PEXtk_MTX_SIZE)) == NULL)
	return 1;
    PEXtk_matrix_copy( mat_stk[p_state->mtx_loc], ident_mat);

    /* Initialize the matrix stacks */

    PEXtk_matrix_copy( t,  ident_mat);
    PEXtk_matrix_copy( s,  ident_mat);
    PEXtk_matrix_copy( rx, ident_mat);
    PEXtk_matrix_copy( ry, ident_mat);
    PEXtk_matrix_copy( rz, ident_mat);

    p_state->mat_stk = mat_stk;

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pget_identity_matrix   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Copies a the 4x4 identity matrix into mat. Immediate mode only.
 *
 *  USAGE:
 *	Pget_identity_matrix( mat )
 *	Matrix	    mat		(out)	Matrix to copy identity matrix to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_identity_matrix( mat )
Matrix	mat;
{
    PEXtk_matrix_copy( mat, ident_mat);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pget_inverse_matrix    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Gets the inverse of the current transformation matrix.
 *
 *  USAGE:
 *	Pget_inverse_matrix( mat )
 *	Matrix	    mat		(in)	Returned matrix
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_inverse_matrix(inv_mat)
Matrix	inv_mat;
{
    register int       i;
    register int       j;
    Matrix		tmp_mat;

	/*  copy the original matrix into a temporary area, set the 
	 *  inverse equal to the identity matrix
	 */

    Pget_matrix(tmp_mat);
    Pget_identity_matrix(inv_mat);


	/* loop through the rows and process --
	 *  + pivot matrix to bring the largest diagonal term (max_piv) into 
	 *      the current row (complete pivoting used)
	 *  + divide the row by the diagonal term - sets the diagonal term
	 *      equal to 0.0.
	 *  + loop through the other rows and add the appropriate factor
	 *      of this row to zero the column of the diagonal
	 */
    {   
	int		k;
	float		f_tmp;
	float		max_piv;

        for (i=0; i<4; i++)
	{
            if (tmp_mat[i][i] < 0.0) max_piv = -tmp_mat[i][i];
            else max_piv = tmp_mat[i][i];
            for (j=i+1, k=i; j<4; j++){
                if (tmp_mat[j][i] > max_piv){
                    max_piv = tmp_mat[j][i];
                    k = j;
                } else if (-tmp_mat[j][i] > max_piv){
                    max_piv = -tmp_mat[j][i];
                    k = j;
                }
            }

	    if (tmp_mat[k][i] == 0.0)             /* singular matrix */
                return ;

	    if (k != i){
		for (j=0; j<4; j++){
		    f_tmp = tmp_mat[i][j];
		    tmp_mat[i][j] = tmp_mat[k][j];
		    tmp_mat[k][j] = f_tmp;
		    f_tmp = inv_mat[i][j];
		    inv_mat[i][j] = inv_mat[k][j];
		    inv_mat[k][j] = f_tmp;
	        }
	    }

	    /*
	     *  factor by inverse of current diagonal value
	     */
	    f_tmp = 1.0 / tmp_mat[i][i];

	    for (j=0; j<4; j++){
		tmp_mat[i][j] *= f_tmp;
		inv_mat[i][j] *= f_tmp;
	    }

	    for (j=0; j<4; j++){
		if (j != i){
		    f_tmp = -tmp_mat[j][i];
		    for(k=0; k<4; k++){
			tmp_mat[j][k] += tmp_mat[i][k] * f_tmp;
			inv_mat[j][k] += inv_mat[i][k] * f_tmp;
		    }
		}
	    }
        } /* end for i */
    } /* end */

    return ;
}

print_mat(mat)
Matrix mat;
{
    int i;
    for(i=0;i<4;i++)
    printf(" %f %f %f %f\n",
        mat[i][0], mat[i][1], mat[i][2], mat[i][3]);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pget_matrix      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Gets the stack's top 4x4 transformation matrix
 *
 *  USAGE:
 *	Pget_matrix( mat )
 *	Matrix	    mat		(in)	matrix to receive xform
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_matrix(mat)
Matrix	mat;
{
    if( p_state->mat_mode == PEXtk_MATRIX_VIEWING )
        PEXtk_matrix_tcopy( mat, p_state->vue->orientation );
    else
    if( p_state->mat_mode == PEXtk_MATRIX_PROJ )
        PEXtk_matrix_tcopy( mat, p_state->vue->mapping );
    else
        PEXtk_matrix_tcopy( mat, mat_stk[p_state->mtx_loc] );
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pload_identity      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Loads a 4x4 identity matrix onto the matrix stack.
 *
 *  USAGE:
 *	Pload_identity( )
 *
 *  NOTE:
 *	This routine can not be used inside a structure for matrix mode
 *	PEXtk_MATRIX_VIEWING or PEXtk_MATRIX_PROJ.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pload_identity()
{
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_LD_IDENT );
	return;
    }
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	    if( p_state->mat_mode == PEXtk_MATRIX_VIEWING )
		PEXtk_matrix_copy( p_state->vue->orientation, ident_mat);
	    else if( p_state->mat_mode == PEXtk_MATRIX_PROJ )
		PEXtk_matrix_copy( p_state->vue->mapping, ident_mat);
	    else {
		PEXtk_matrix_copy( mat_stk[p_state->mtx_loc], ident_mat);
		p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
		PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
			mat_stk[p_state->mtx_loc]);
		memcpy( p_state->pcAttrib.global_transform,
			mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
		PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	    }
    } else {
	PEXtk_matrix_copy( local_mtx, ident_mat);
	/* composition = PEXReplace ; */
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type, 
		PEXReplace, local_mtx);
    }

}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pload_matrix      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Copies a 4x4 transformation matrix onto the matrix stack.
 *
 *  USAGE:
 *	Pload_matrix( mat )
 *	Matrix	    mat		(in)	matrix to load
 *
 *  NOTE:
 *	This routine can not be used inside a structure for matrix mode
 *	PEXtk_MATRIX_VIEWING or PEXtk_MATRIX_PROJ.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pload_matrix(mat)
Matrix	mat;
{
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_matrix( PEXtk_OP_LD_MTX, mat, 0, 0);
	return;
    }
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	    if( p_state->mat_mode == PEXtk_MATRIX_VIEWING )
		PEXtk_matrix_tcopy( p_state->vue->orientation, mat);
	    else if( p_state->mat_mode == PEXtk_MATRIX_PROJ )
		PEXtk_matrix_tcopy( p_state->vue->mapping, mat);
	    else {
		PEXtk_matrix_tcopy( mat_stk[p_state->mtx_loc], mat);
		p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
		PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
			mat_stk[p_state->mtx_loc]);
		memcpy( p_state->pcAttrib.global_transform,
			mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
		PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	    }
    } else {
	PEXtk_matrix_tcopy( local_mtx, mat);
	/* composition = PEXReplace; */
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type, 
		PEXReplace, local_mtx);
    }

#ifdef PRINT_STK
{ int i; printf(" Matrix\n");
for(i=0;i<4;i++)
printf(" %f %f %f %f\n",
	mat_stk[p_state->mtx_loc][i][0], mat_stk[p_state->mtx_loc][i][1],
	mat_stk[p_state->mtx_loc][i][2], mat_stk[p_state->mtx_loc][i][3]);
}
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pmultiply_matrix     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Multiplies a 4x4 matrix with the top of the matrix stack.
 *
 *  USAGE:
 *	Pmultiply_matrix( mat, flag )
 *	Matrix	    mat		(in)	matrix to load
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmultiply_matrix(mat, flag)
Matrix	mat;
int	flag;
{
    int composition;
    Matrix prod;
    float *ptr;

    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	/* DLS_matrix( PEXtk_OP_MULT_MTX, mat, flag, 0); */
	return;
    }

    if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	ptr = &mat_stk[p_state->mtx_loc][0][0];
    else
	ptr = &p_state->vue->orientation[0][0];
    
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( POSTFLAG_IS_TRUE(flag) )
	    PEXtk_matrix_mult( mat, ptr, prod);
	else
	    PEXtk_matrix_mult( ptr, mat, prod);
	PEXtk_matrix_tcopy( ptr, prod);

	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}

    } else {
	PEXtk_matrix_tcopy( local_mtx, mat);
	if( POSTFLAG_IS_TRUE(flag) )
	    composition = PEXPostConcatenate;
	else
	    composition = PEXPreConcatenate;
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }

}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Ppop_matrix       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pops the matrix stack.
 *
 *  USAGE:
 *	Ppop_matrix( )
 *
 *  NOTE:
 *	This routine is for Immediate Mode only.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppop_matrix( )
{
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    if( p_state->mtx_loc > 0 )
		p_state->mtx_loc--;
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    if( p_state->view_loc > 0 )
		p_state->view_loc--;
            p_state->view = &p_state->view_stk[p_state->view_loc];
            p_state->vue  = &p_state->vue_tbl[p_state->view_loc];
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}
	return;
    } else
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_stack( PEXtk_OP_POP_MTX);
	return;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Ppush_matrix      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pushes the matrix stack.
 *
 *  USAGE:
 *	Ppush_matrix( )
 *
 *  NOTE:
 *	This routine is for Immediate Mode only.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppush_matrix( )
{
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    if( p_state->mtx_loc < PEXtk_MTX_SIZE-1 ){
		p_state->mtx_loc++;
		PEXtk_matrix_copy(mat_stk[p_state->mtx_loc], 
		    mat_stk[p_state->mtx_loc-1]);
	    }
	} else {
	    if( p_state->view_loc < PEXtk_VIEW_SIZE-1 ){
		p_state->view_loc++;
        	p_state->view = &p_state->view_stk[p_state->view_loc];
        	p_state->vue  = &p_state->vue_tbl[p_state->view_loc];
		memcpy( (char *) &p_state->view_stk[p_state->view_loc], 
		        (char *) &p_state->view_stk[p_state->view_loc-1], 
			sizeof(PEXtk_VIEW));
		memcpy( (char *) &p_state->vue_tbl[p_state->view_loc],
		        (char *) &p_state->vue_tbl[p_state->view_loc-1], 
			sizeof(PEXViewEntry));

		PEXSetTableEntries( dpy, 
		    p_state->win[p_state->cur_w].View_ID, 
		    p_state->view_loc, 1, PEXLUTView, p_state->vue);
		PEXtk_reset_pex_state( 0, PEXRAViewTable);
	    }
	}
	return;
    } else
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_stack( PEXtk_OP_PUSH_MTX);
	return;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Protate          +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the x,y, or z rotation matrix and concats it to top of stack
 *
 *  USAGE:
 *	Protate( a, axis, flag )
 *	Angle	    a		(in)	Angle, in degrees
 *	char	    axis	(in)	Rotation axis, 'x', 'y', or 'z'
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Protate( a, axis, flag)
Angle	a;
char	axis;
int	flag;
{

	switch( axis ){
	  case 'x':
		Protate_x( a, flag);
		break;
	  case 'y':
		Protate_y( a, flag);
		break;
	  case 'z':
		Protate_z( a, flag);
		break;
	}
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Protate_vector      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the rotation matrix for rotation about a vector 
 *	and concatenates it to top of stack
 *
 *  USAGE:
 *	Protate_vector( ox, oy, oz, vx, vy, vz, a, flag )
 *	Fcoord	    ox		(in)	Vector start x value
 *	Fcoord	    oy		(in)	Vector start y value
 *	Fcoord	    oz		(in)	Vector start z value
 *	Fcoord	    vx		(in)	Vector end x value
 *	Fcoord	    vy		(in)	Vector end y value
 *	Fcoord	    vz		(in)	Vector end z value
 *	Angle	    a		(in)	Angle, in degrees
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *  NOTE:
 *	NOT IMPLEMENTED YET
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Protate_vector( ox, oy, oz, vx, vy, vz, a, flag )
Fcoord	    ox;
Fcoord	    oy;
Fcoord	    oz;
Fcoord	    vx;
Fcoord	    vy;
Fcoord	    vz;
Angle	    a;
int	    flag;
{
    Matrix mat, prod;
    float  *ptr;
    double rad;
    float  dx, dy, dz, rx, ry, rz, rx2, ry2, rz2;
    float  dist, sav1, sav2, cosine, sine, cosf1;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	
	/* Calculate the distance between 'o' and 'v' */
	dx = vx - ox;
	dy = vy - oy;
	dz = vz - oz;

	dist = (float) sqrt( (double)( dx*dx + dy*dy + dz*dz));

	/* Calculate the normalized rotation vector */
	rx = dx/dist;
	ry = dy/dist;
	rz = dz/dist;

	rx2 = rx*rx;
	ry2 = ry*ry;
	rz2 = rz*rz;

	/* Calculate the required trig functions */
	rad = a * 3.141592653589793 / 180.0;
	cosine = cos(rad);
	sine   = sin(rad);
	cosf1  = 1.0 - cosine;

	/* Calculate the diagonal terms */
	mat[0][0] = rx2 + cosine * (1.0 - rx2);
	mat[1][1] = ry2 + cosine * (1.0 - ry2);
	mat[2][2] = rz2 + cosine * (1.0 - rz2);

	/* Calculate the cross x:y terms */
	sav1 = rx*ry*cosf1;
	sav2 = rz*sine;
	mat[1][0] = sav1 - sav2;
	mat[0][1] = sav1 + sav2;

	/* Calculate the cross x:z terms */
	sav1 = rx*rz*cosf1;
	sav2 = ry*sine;
	mat[2][0] = sav1 - sav2;
	mat[0][2] = sav1 + sav2;

	/* Calculate the cross y:z terms */
	sav1 = ry*rz*cosf1;
	sav2 = rx*sine;
	mat[2][1] = sav1 - sav2;
	mat[1][2] = sav1 + sav2;

	ptr = &mat[0][0];
	if( ox != 0.0 ){
	    *(ptr)   = *(ptr) + *(ptr+12)*ox;
	    *(ptr+1) = *(ptr+1) + *(ptr+13)*ox;
	    *(ptr+2) = *(ptr+2) + *(ptr+14)*ox;
	    *(ptr+3) = *(ptr+3) + *(ptr+15)*ox;
	}

	if( oy != 0.0 ){
	    *(ptr+4) = *(ptr+4) + *(ptr+12)*oy;
	    *(ptr+5) = *(ptr+5) + *(ptr+13)*oy;
	    *(ptr+6) = *(ptr+6) + *(ptr+14)*oy;
	    *(ptr+7) = *(ptr+7) + *(ptr+15)*oy;
	}

	if( oz != 0.0 ){
	    *(ptr+8) = *(ptr+8) + *(ptr+12)*oz;
	    *(ptr+9) = *(ptr+9) + *(ptr+13)*oz;
	    *(ptr+10) = *(ptr+10) + *(ptr+14)*oz;
	    *(ptr+11) = *(ptr+11) + *(ptr+15)*oz;
	}

	*(ptr+3)  = *ptr*(-ox) - *(ptr+1)*oy - *(ptr+2)*oz + *(ptr+3);
	*(ptr+7)  = *(ptr+4)*(-ox) - *(ptr+5)*oy - *(ptr+6)*oz + *(ptr+7);
	*(ptr+11) = *(ptr+8)*(-ox) - *(ptr+9)*oy - *(ptr+10)*oz + *(ptr+11);
	*(ptr+15) = *(ptr+12)*(-ox) - *(ptr+13)*oy - *(ptr+14)*oz + *(ptr+15);

    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( POSTFLAG_IS_TRUE(flag) )	/* postconcat */
	    PEXtk_matrix_mult( mat, mat_stk[p_state->mtx_loc], prod);
	else		/* preconcat */
	    PEXtk_matrix_mult( mat_stk[p_state->mtx_loc], mat, prod);
	p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type, prod);
	memcpy( p_state->pcAttrib.global_transform, prod, sizeof(PEXMatrix));
	PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	PEXtk_matrix_copy( mat_stk[p_state->mtx_loc], prod);
    }

    return;

}
/*
 *
 *                       +--------------------------+
 * ----------------------+        Protate_x         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the x rotation matrix and concatenates it to top of stack
 *
 *  USAGE:
 *	Protate_x( a, flag )
 *	Angle	    a		(in)	Angle, in degrees
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  NOTE:
 *	This routine needs to be optimized to only multiply those
 *	rows/cols that are affected by a rotate x.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Protate_x( a, flag)
Angle	a;
int	flag;
{
    int composition;
    double rad;
    float *ptr;
    register float *ptr1, *ptr2;
    float sav[8];
    float crad, srad;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    rad = a * 3.141592653589793 / 180.0;

    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

	crad = (float) cos(rad);
	srad = (float) sin(rad);
	if( POSTFLAG_IS_TRUE(flag) ){		/* PostConcatenate */
	    ptr1 = ptr + 4; ptr2 = ptr + 8;
	    sav[0] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[1] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 5; ptr2 = ptr + 9;
	    sav[2] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[3] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 6; ptr2 = ptr + 10;
	    sav[4] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[5] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 7; ptr2 = ptr + 11;
	    sav[6] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[7] = *(ptr1)*srad + *(ptr2)*crad;

	    *(ptr+4) = sav[0];  
	    *(ptr+8) = sav[1];
	    *(ptr+5) = sav[2];
	    *(ptr+9) = sav[3];
	    *(ptr+6) = sav[4];
	    *(ptr+10) = sav[5];
	    *(ptr+7) = sav[6];
	    *(ptr+11) = sav[7];
	} else {				/* PreConcatenate */
	    ptr1 = ptr + 1; ptr2 = ptr + 2;
	    sav[0] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[1] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 5; ptr2 = ptr + 6;
	    sav[2] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[3] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 9; ptr2 = ptr + 10;
	    sav[4] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[5] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 13; ptr2 = ptr + 14;
	    sav[6] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[7] = *(ptr2)*crad - *(ptr1)*srad;

	    *(ptr+1) = sav[0]; 
	    *(ptr+2) = sav[1];
	    *(ptr+5) = sav[2];
	    *(ptr+6) = sav[3];
	    *(ptr+9) = sav[4];
	    *(ptr+10) = sav[5];
	    *(ptr+13) = sav[6];
	    *(ptr+14) = sav[7];
	}
		
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}

    } else if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_xform( PEXtk_OP_ROT_X, flag, a, 0.0, 0.0);
    } else {
	rx[1][1] = cos(rad);
	rx[2][1] = sin(rad);
	rx[1][2] = -rx[2][1];
	rx[2][2] = rx[1][1];
	PEXtk_matrix_copy( local_mtx, rx);
	if( POSTFLAG_IS_TRUE(flag) ){
	    composition = PEXPostConcatenate;
	} else {
	    composition = PEXPreConcatenate;
	}
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }

}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Protate_y         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the y rotation matrix and concatenates it to top of stack
 *
 *  USAGE:
 *	Protate_y( a, flag )
 *	Angle	    a		(in)	Angle, in degrees
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  NOTE:
 *	This routine needs to be optimized to only multiply those
 *	rows/cols that are affected by a rotate y.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Protate_y( a, flag)
Angle	a;
int	flag;
{
    int composition;
    double rad;
    float *ptr;
    register float *ptr1, *ptr2;
    float sav[8];
    float crad, srad;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    rad = a * 3.141592653589793 / 180.0;

    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

	crad = (float) cos(rad);
	srad = (float) sin(rad);
	if( POSTFLAG_IS_TRUE(flag) ){		/* PostConcatenate */
	    ptr1 = ptr ; ptr2 = ptr + 8;
	    sav[0] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[1] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 1; ptr2 = ptr + 9;
	    sav[2] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[3] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 2; ptr2 = ptr + 10;
	    sav[4] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[5] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 3; ptr2 = ptr + 11;
	    sav[6] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[7] = *(ptr2)*crad - *(ptr1)*srad;

	    *(ptr)   = sav[0];  
	    *(ptr+8) = sav[1];
	    *(ptr+1) = sav[2];
	    *(ptr+9) = sav[3];
	    *(ptr+2) = sav[4];
	    *(ptr+10) = sav[5];
	    *(ptr+3) = sav[6];
	    *(ptr+11) = sav[7];
	} else {				/* PreConcatenate */
	    ptr1 = ptr ; ptr2 = ptr + 2;
	    sav[0] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[1] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 4; ptr2 = ptr + 6;
	    sav[2] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[3] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 8; ptr2 = ptr + 10;
	    sav[4] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[5] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 12; ptr2 = ptr + 14;
	    sav[6] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[7] = *(ptr1)*srad + *(ptr2)*crad;

	    *(ptr)   = sav[0]; 
	    *(ptr+2) = sav[1];
	    *(ptr+4) = sav[2];
	    *(ptr+6) = sav[3];
	    *(ptr+8) = sav[4];
	    *(ptr+10) = sav[5];
	    *(ptr+12) = sav[6];
	    *(ptr+14) = sav[7];
	}
		
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}
		
    } else if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_xform( PEXtk_OP_ROT_Y, flag, a, 0.0, 0.0);
    } else {
	ry[0][0] = cos(rad);
	ry[0][2] = sin(rad);
	ry[2][0] = -ry[0][2];
	ry[2][2] = ry[0][0];
	PEXtk_matrix_copy( local_mtx, ry);
	if( POSTFLAG_IS_TRUE(flag) ){
	    composition = PEXPostConcatenate;
	} else {
	    composition = PEXPreConcatenate;
	}
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Protate_z         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the z rotation matrix and concatenates it to top of stack
 *
 *  USAGE:
 *	Protate_z( a, flag )
 *	Angle	    a		(in)	Angle, in degrees
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *  NOTE:
 *	This routine needs to be optimized to only multiply those
 *	rows/cols that are affected by a rotate z.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Protate_z( a, flag)
Angle	a;
int	flag;
{
    int composition;
    double rad;
    float *ptr;
    register float *ptr1, *ptr2;
    float sav[8];
    float crad, srad;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    rad = a * 3.141592653589793 / 180.0;

    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

	crad = (float) cos(rad);
	srad = (float) sin(rad);
	if( POSTFLAG_IS_TRUE(flag) ){		/* PostConcatenate */
	    ptr1 = ptr ; ptr2 = ptr + 4;
	    sav[0] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[1] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 1; ptr2 = ptr + 5;
	    sav[2] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[3] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 2; ptr2 = ptr + 6;
	    sav[4] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[5] = *(ptr1)*srad + *(ptr2)*crad;
	    ptr1 = ptr + 3; ptr2 = ptr + 7;
	    sav[6] = *(ptr1)*crad - *(ptr2)*srad;
	    sav[7] = *(ptr1)*srad + *(ptr2)*crad;

	    *(ptr)   = sav[0];  
	    *(ptr+4) = sav[1];
	    *(ptr+1) = sav[2];
	    *(ptr+5) = sav[3];
	    *(ptr+2) = sav[4];
	    *(ptr+6) = sav[5];
	    *(ptr+3) = sav[6];
	    *(ptr+7) = sav[7];
	} else {				/* PreConcatenate */
	    ptr1 = ptr ; ptr2 = ptr + 1;
	    sav[0] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[1] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 4; ptr2 = ptr + 5;
	    sav[2] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[3] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 8; ptr2 = ptr + 9;
	    sav[4] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[5] = *(ptr2)*crad - *(ptr1)*srad;
	    ptr1 = ptr + 12; ptr2 = ptr + 13;
	    sav[6] = *(ptr1)*crad + *(ptr2)*srad;
	    sav[7] = *(ptr2)*crad - *(ptr1)*srad;

	    *(ptr)   = sav[0]; 
	    *(ptr+1) = sav[1];
	    *(ptr+4) = sav[2];
	    *(ptr+5) = sav[3];
	    *(ptr+8) = sav[4];
	    *(ptr+9) = sav[5];
	    *(ptr+12) = sav[6];
	    *(ptr+13) = sav[7];
	}
		
	/* Here is a place to optimize... 
	 *     Don't send down the matrix until geometry is invoked
	 */
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}

    } else if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_xform( PEXtk_OP_ROT_Z, flag, a, 0.0, 0.0);
    } else {
	rz[0][0] = cos(rad);
	rz[1][0] = sin(rad);
	rz[0][1] = -rz[1][0];
	rz[1][1] = rz[0][0];
	PEXtk_matrix_copy( local_mtx, rz);
	if( POSTFLAG_IS_TRUE(flag) ){
	    composition = PEXPostConcatenate;
	} else {
	    composition = PEXPreConcatenate;
	}
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Pscale          +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the x,y,z scaling matrix and concatenates it to top of stack
 *
 *  USAGE:
 *	Pscale( x, y, z, flag )
 *	Fcoord	    x		(in)	x scale
 *	Fcoord	    y		(in)	y scale
 *	Fcoord	    z		(in)	z scale
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pscale( x, y, z, flag)
Fcoord	x;
Fcoord	y;
Fcoord	z;
int	flag;
{

    int composition;
    float *ptr;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

    if( POSTFLAG_IS_TRUE(flag) ){		/* PostConcatenate */
	if( x != 1.0 ){
	    *(ptr)    = *(ptr)*x;
	    *(ptr+4)  = *(ptr+4)*x;
	    *(ptr+8)  = *(ptr+8)*x;
	    *(ptr+12) = *(ptr+12)*x;
	}
	if( y != 1.0 ){
	    *(ptr+1)  = *(ptr+1)*y;
	    *(ptr+5)  = *(ptr+5)*y;
	    *(ptr+9)  = *(ptr+9)*y;
	    *(ptr+13) = *(ptr+13)*y;
	}
	if( z != 1.0 ){
	    *(ptr+2)  = *(ptr+2)*z;
	    *(ptr+6)  = *(ptr+6)*z;
	    *(ptr+10) = *(ptr+10)*z;
	    *(ptr+14) = *(ptr+14)*z;
	}
    } else {					/* PreConcatenate */
	if( x != 1.0 ){
	    *(ptr)    = *(ptr)*x;
	    *(ptr+1)  = *(ptr+1)*x;
	    *(ptr+2)  = *(ptr+2)*x;
	    *(ptr+3)  = *(ptr+3)*x;
	}
	if( y != 1.0 ){
	    *(ptr+4)  = *(ptr+4)*y;
	    *(ptr+5)  = *(ptr+5)*y;
	    *(ptr+6)  = *(ptr+6)*y;
	    *(ptr+7)  = *(ptr+7)*y;
	}
	if( z != 1.0 ){
	    *(ptr+8)  = *(ptr+8)*z;
	    *(ptr+9)  = *(ptr+9)*z;
	    *(ptr+10) = *(ptr+10)*z;
	    *(ptr+11) = *(ptr+11)*z;
	}
    }
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}

    } else if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_xform( PEXtk_OP_SCALE, flag, x, y, z);
    } else {
	s[0][0] = x;
	s[1][1] = y;
	s[2][2] = z;
	PEXtk_matrix_copy( local_mtx, s);
	if( POSTFLAG_IS_TRUE(flag) ){
	    composition = PEXPostConcatenate;
	} else {
	    composition = PEXPreConcatenate;
	}
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Ptranslate        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the x,y,z translation matrix and concatenates it 
 *	to the top of stack
 *
 *  USAGE:
 *	Ptranslate( x, y, z, flag )
 *	Fcoord	    x		(in)	x translation
 *	Fcoord	    y		(in)	x translation
 *	Fcoord	    z		(in)	x translation
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  NOTE:
 *	This routine needs to be optimized to only multiply those
 *	rows/cols that are affected by translate x,y,z.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptranslate( x, y, z, flag)
Fcoord	x;
Fcoord	y;
Fcoord	z;
int	flag;
{
    int composition;
    float *ptr;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

      if( POSTFLAG_IS_TRUE(flag) ){ 	/* postconcat */
	if( x != 0.0 ){
	    *(ptr)   = *(ptr) + *(ptr+12)*x;
	    *(ptr+1) = *(ptr+1) + *(ptr+13)*x;
	    *(ptr+2) = *(ptr+2) + *(ptr+14)*x;
	    *(ptr+3) = *(ptr+3) + *(ptr+15)*x;
	}

	if( y != 0.0 ){
	    *(ptr+4) = *(ptr+4) + *(ptr+12)*y;
	    *(ptr+5) = *(ptr+5) + *(ptr+13)*y;
	    *(ptr+6) = *(ptr+6) + *(ptr+14)*y;
	    *(ptr+7) = *(ptr+7) + *(ptr+15)*y;
	}

	if( z != 0.0 ){
	    *(ptr+8) = *(ptr+8) + *(ptr+12)*z;
	    *(ptr+9) = *(ptr+9) + *(ptr+13)*z;
	    *(ptr+10) = *(ptr+10) + *(ptr+14)*z;
	    *(ptr+11) = *(ptr+11) + *(ptr+15)*z;
	}
      } else {		/* preconcat */
	*(ptr+3)  = *ptr*x + *(ptr+1)*y + *(ptr+2)*z + *(ptr+3);
	*(ptr+7)  = *(ptr+4)*x + *(ptr+5)*y + *(ptr+6)*z + *(ptr+7);
	*(ptr+11) = *(ptr+8)*x + *(ptr+9)*y + *(ptr+10)*z + *(ptr+11);
	*(ptr+15) = *(ptr+12)*x + *(ptr+13)*y + *(ptr+14)*z + *(ptr+15);
      }
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}

    } else if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_xform( PEXtk_OP_TRAN, flag, x, y, z);
    } else {
	t[0][3] = x;
	t[1][3] = y;
	t[2][3] = z;
	PEXtk_matrix_copy( local_mtx, t);
	if( POSTFLAG_IS_TRUE(flag) ){
	    composition = PEXPostConcatenate;
	} else {
	    composition = PEXPreConcatenate;
	}
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }

}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ptranslate_x       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the x translation matrix and concatenates it 
 *	to the top of stack
 *
 *  USAGE:
 *	Ptranslate_x( x, flag )
 *	Fcoord	    x		(in)	x translation
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  NOTE:
 *	This routine needs to be optimized to only multiply those
 *	rows/cols that are affected by translate x.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptranslate_x( x, flag)
Fcoord	x;
int	flag;
{
    int composition;
    float *ptr;
    float sav1, sav2, sav3, sav4;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

	if( POSTFLAG_IS_TRUE(flag) ){	/* postconcat */
	    sav1 = *(ptr) + *(ptr+12)*x;
	    sav2 = *(ptr+1) + *(ptr+13)*x;
	    sav3 = *(ptr+2) + *(ptr+14)*x;
	    sav4 = *(ptr+3) + *(ptr+15)*x;

	    *(ptr)   = sav1; /* last col */
	    *(ptr+1) = sav2;
	    *(ptr+2) = sav3;
	    *(ptr+3) = sav4;
	} else {	/* preconcat */
	    sav1 = *(ptr)*x + *(ptr+3);
	    sav2 = *(ptr+4)*x + *(ptr+7);
	    sav3 = *(ptr+8)*x + *(ptr+11);
	    sav4 = *(ptr+12)*x + *(ptr+15);

	    *(ptr+3) = sav1;  /* last column */
	    *(ptr+7) = sav2;
	    *(ptr+11) = sav3;
	    *(ptr+15) = sav4;
	}
		
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}

    } else if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_xform( PEXtk_OP_TRAN_X, flag, x, 0.0, 0.0);
    } else {
	t[0][3] = x;
	t[1][3] = 0.0;
	t[2][3] = 0.0;
	PEXtk_matrix_copy( local_mtx, t);
	if( POSTFLAG_IS_TRUE(flag) ){
	    composition = PEXPostConcatenate;
	} else {
	    composition = PEXPreConcatenate;
	}
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }

}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ptranslate_y       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the y translation matrix and concatenates it 
 *	to the top of stack
 *
 *  USAGE:
 *	Ptranslate_y( y, flag )
 *	Fcoord	    y		(in)	x translation
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  NOTE:
 *	This routine needs to be optimized to only multiply those
 *	rows/cols that are affected by translate y.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptranslate_y( y, flag)
Fcoord	y;
int	flag;
{
    int composition;
    float *ptr;
    float sav1, sav2, sav3, sav4;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

	if( POSTFLAG_IS_TRUE(flag) ){	/* postconcat */
	    sav1 = *(ptr+4) + *(ptr+12)*y;
	    sav2 = *(ptr+5) + *(ptr+13)*y;
	    sav3 = *(ptr+6) + *(ptr+14)*y;
	    sav4 = *(ptr+7) + *(ptr+15)*y;

	    *(ptr+4) = sav1; /* last col */
	    *(ptr+5) = sav2;
	    *(ptr+6) = sav3;
	    *(ptr+7) = sav4;
	} else {	/* preconcat */
	    sav1 = *(ptr+1)*y + *(ptr+3);
	    sav2 = *(ptr+5)*y + *(ptr+7);
	    sav3 = *(ptr+9)*y + *(ptr+11);
	    sav4 = *(ptr+13)*y + *(ptr+15);

	    *(ptr+3) = sav1;  /* last column */
	    *(ptr+7) = sav2;
	    *(ptr+11) = sav3;
	    *(ptr+15) = sav4;
	}
		
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}

    } else if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_xform( PEXtk_OP_TRAN_Y, flag, y, 0.0, 0.0);
    } else {
	t[0][3] = 0.0;
	t[1][3] = y;
	t[2][3] = 0.0;
	PEXtk_matrix_copy( local_mtx, t);
	if( POSTFLAG_IS_TRUE(flag) ){
	    composition = PEXPostConcatenate;
	} else {
	    composition = PEXPreConcatenate;
	}
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ptranslate_z       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Calculates the z translation matrix and concatenates it 
 *	to the top of stack
 *
 *  USAGE:
 *	Ptranslate_z( z, flag )
 *	Fcoord	    z		(in)	x translation
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  NOTE:
 *	This routine needs to be optimized to only multiply those
 *	rows/cols that are affected by translate z.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptranslate_z( z, flag)
Fcoord	z;
int	flag;
{

    int composition;
    float *ptr;
    float sav1, sav2, sav3, sav4;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

	if( POSTFLAG_IS_TRUE(flag) ){	/* postconcat */
	    sav1 = *(ptr+8) + *(ptr+12)*z;
	    sav2 = *(ptr+9) + *(ptr+13)*z;
	    sav3 = *(ptr+10) + *(ptr+14)*z;
	    sav4 = *(ptr+11) + *(ptr+15)*z;

	    *(ptr+8) = sav1; /* last col */
	    *(ptr+9) = sav2;
	    *(ptr+10) = sav3;
	    *(ptr+11) = sav4;
	} else {	/* preconcat */
	    sav1 = *(ptr+2)*z + *(ptr+3);
	    sav2 = *(ptr+6)*z + *(ptr+7);
	    sav3 = *(ptr+10)*z + *(ptr+11);
	    sav4 = *(ptr+14)*z + *(ptr+15);

	    *(ptr+3) = sav1;  /* last column */
	    *(ptr+7) = sav2;
	    *(ptr+11) = sav3;
	    *(ptr+15) = sav4;
	}
		
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING ){
	    p_state->dirty_mask &= PEXtk_DIRTY_MTX;
#ifdef GLOBAL_XFORM
	    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
		mat_stk[p_state->mtx_loc]);
	    memcpy( p_state->pcAttrib.global_transform,
		mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
#endif
	} else {
	    PEXSetTableEntries( dpy, 
		p_state->win[p_state->cur_w].View_ID, 
		p_state->view_loc, 1, PEXLUTView, p_state->vue);
	    PEXtk_reset_pex_state( 0, PEXRAViewTable);
	}

    } else if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_xform( PEXtk_OP_TRAN_Z, flag, z, 0.0, 0.0);
    } else {
	t[0][3] = 0.0;
	t[1][3] = 0.0;
	t[2][3] = z;
	PEXtk_matrix_copy( local_mtx, t);
	if( POSTFLAG_IS_TRUE(flag) ){
	    composition = PEXPostConcatenate;
	} else {
	    composition = PEXPreConcatenate;
	}
	PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	    composition, local_mtx);
    }
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_Update_Matrix   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Updates the Global matrix with the top of the matrix stack.
 *
 *  USAGE:
 *	PEXtk_Update_Matrix( flag )
 *	int 	    flag	(in)	Flag to send OC 
 *					    0 - Don't set global transform OC
 *					    1 - Set global transform OC
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_Update_Matrix( flag )
int	flag;
{
    if( flag )
	PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
	    mat_stk[p_state->mtx_loc]);
    memcpy( p_state->pcAttrib.global_transform,
	mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_preconcat     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Premultiplies the top of the matrix stack with an input matrix.
 *
 *  USAGE:
 *	PEXtk_preconcat( mat )
 *	Matrix	    mat		(in)	Matrix to multiply stack with
 *
 *  NOTE:
 *	This routine is outdated.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_preconcat( mat )
Matrix mat;
{
    Matrix prod;

    PEXtk_matrix_mult( mat_stk[p_state->mtx_loc], mat, prod);
    PEXtk_matrix_copy(mat_stk[p_state->mtx_loc], prod);
    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type,
	mat_stk[p_state->mtx_loc]);
    memcpy( p_state->pcAttrib.global_transform,
	mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_preconcat_struct  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Premultiplies the top of the matrix stack with an input matrix.
 *	This routine is intended to be used inside a structure.
 *
 *  USAGE:
 *	PEXtk_preconcat_struct( mat )
 *	Matrix	    mat		(in)	Matrix to multiply stack with
 *
 *  NOTE:
 *	This routine is the concat routine for PEXtk matrices. Since this
 *	routine is only called inside PEXtk, the input matrix does not
 *	have to be transposed.
 *
 *  NOTE:
 *	This routine is outdated.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_preconcat_struct( mat )
Matrix mat;
{
    int composition;

    /* PEXtk_matrix_tcopy( local_mtx, mat); */
    PEXtk_matrix_copy( local_mtx, mat);
    composition = PEXPreConcatenate;
    PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type,
	composition, local_mtx);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       PEXtk_concat       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Postmultiplies the top of the matrix stack with an input matrix.
 *
 *  USAGE:
 *	PEXtk_concat( mat )
 *	Matrix	    mat		(in)	Matrix to multiply stack with
 *
 *  NOTE:
 *	This routine is outdated.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_concat( mat )
Matrix mat;
{
    Matrix prod;

    PEXtk_matrix_mult( mat, mat_stk[p_state->mtx_loc], prod);
    PEXtk_matrix_copy(mat_stk[p_state->mtx_loc], prod);
    PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type, 
	mat_stk[p_state->mtx_loc]);
    memcpy( p_state->pcAttrib.global_transform,
	mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
    PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_concat_struct   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Postmultiplies the top of the matrix stack with an input matrix.
 *	This routine is intended to be used inside a structure.
 *
 *  USAGE:
 *	PEXtk_concat( mat )
 *	Matrix	    mat		(in)	Matrix to multiply stack with
 *
 *  NOTE:
 *	This routine is the concat routine for PEXtk matrices. Since this
 *	routine is only called inside PEXtk, the input matrix does not
 *	have to be transposed.
 *
 *  NOTE:
 *	This routine is outdated.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_concat_struct( mat )
Matrix mat;
{
    int composition;

    /* PEXtk_matrix_tcopy( local_mtx, mat); */
    PEXtk_matrix_copy( local_mtx, mat);
    composition = PEXPostConcatenate;
    PEXSetLocalTransform( dpy, p_state->cur_id, p_state->req_type, 
	composition, local_mtx);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_matrix_mult     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Support routine. Multiplies two matrices.
 *
 *  USAGE:
 *	PEXtk_matrix_mult( l, r, prod )
 *	Matrix	    l		(in)	left hand side matrix
 *	Matrix	    r		(in)	right hand side matrix
 *	Matrix	    prod	(out)	destination matrix
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_matrix_mult( l, r, prod)
Matrix l;
Matrix r;
Matrix prod;
{
    int i,j;

    for(i = 0; i < 4; i++)
        for(j = 0; j < 4; j++)
            prod[i][j] =      l[i][0] * r[0][j] +
                              l[i][1] * r[1][j] +
                              l[i][2] * r[2][j] +
                              l[i][3] * r[3][j];

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     PEXtk_matrix_copy    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Copies a 4x4 matrix from src to dst
 *
 *  USAGE:
 *	PEXtk_matrix_copy( dst, src )
 *	Matrix	    dst		(out)	destination matrix
 *	Matrix	    src		(in)	source matrix
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_matrix_copy( dst, src)
Matrix dst;
Matrix src;
{
    int i;
    float *dstptr, *srcptr;

    dstptr = (float *)dst;
    srcptr = (float *)src;

    for(i=0;i<16;i++, dstptr++, srcptr++)
	*dstptr = *srcptr;

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     PEXtk_matrix_tcopy   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Copies a transpose of a 4x4 matrix from src to dst
 *
 *  USAGE:
 *	PEXtk_matrix_tcopy( dst, src )
 *	Matrix	    dst		(out)	destination matrix
 *	Matrix	    src		(in)	source matrix. The transpose of this
 *					matrix is copied into dst.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_matrix_tcopy( dst, src)
Matrix	dst;
Matrix	src;
{
    register int i,j;

    for(i=0;i<4;i++)
	for(j=0;j<4;j++)
	{
	    dst[i][j] = src[j][i];
	}

    return 0;
}

